from __future__ import annotations

import glob
import json
import os
import re
from typing import Optional, Union, Iterable

import pandas as pd
from deprecated import deprecated

from seeq import spy
from seeq.base import util
from seeq.sdk import *
from seeq.spy import _common, _login
from seeq.spy import _url
from seeq.spy._common import safe_json_dumps
from seeq.spy._errors import *
from seeq.spy._redaction import safely
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.workbooks._annotation import Journal, Report
from seeq.spy.workbooks._content import DateRange, AssetSelection
from seeq.spy.workbooks._context import WorkbookPushContext
from seeq.spy.workbooks._item import Item, ItemList
from seeq.spy.workbooks._item_map import ItemMap
from seeq.spy.workbooks._table_toolbar import ConditionTableToolbar
from seeq.spy.workbooks._trend_toolbar import TrendToolbar
from seeq.spy.workbooks._workstep import Workstep, AnalysisWorkstep


class Worksheet(Item):
    _annotation: Union[Journal, Report]
    _item_map_for_push: Optional[ItemMap]

    def __new__(cls, *args, **kwargs):
        if cls is Worksheet:
            raise SPyTypeError("Worksheet may not be instantiated directly, create either AnalysisWorksheet or "
                               "TopicWorksheet")

        return object.__new__(cls)

    def __init__(self, workbook, definition=None, *, provenance=None, add_defaults=True):
        super().__init__(definition, provenance=provenance)

        if workbook is None:
            raise SPyValueError("A Workbook is required to create a Worksheet")

        self.workbook = workbook
        self.workbook.worksheets.append(self)

        if add_defaults:
            if self.workbook['Workbook Type'] == 'Analysis':
                self._annotation = Journal(self)
                self.worksteps = dict()
                workstep = AnalysisWorkstep(self)
                self._definition['Current Workstep ID'] = workstep.id
            else:
                self._annotation = Report(self)

    @property
    def url(self):
        """
        The URL of the worksheet. Note that value won't be filled if the workbook/worksheet hasn't been pulled from
        Seeq.
        """
        # Note that 'URL' won't be filled in if a workbook/worksheet hasn't been pushed/pulled. That's because the
        # IDs may change from the placeholders that get generated.
        return self['URL']

    @property
    def fqn(self):
        return f'{self.workbook.fqn} >> {self.name}'

    @property
    @deprecated(reason='Use either journal (for AnalysisWorksheet) or report (for TopicDocument) property instead')
    def document(self):
        return self._annotation

    @staticmethod
    def _instantiate(workbook, definition=None, *, provenance=None):
        if workbook['Workbook Type'] == 'Analysis':
            return AnalysisWorksheet(workbook, definition, provenance=provenance)
        else:
            return TopicDocument(workbook, definition, provenance=provenance)

    @staticmethod
    def pull(item_id, *, workbook=None, extra_workstep_tuples=None, include_annotations=True, include_images=True,
             include_archived: Optional[bool] = None, session: Optional[Session] = None, status=None):
        session = Session.validate(session)
        status = Status.validate(status, session)

        if workbook is None:
            raise SPyValueError('workbook argument is None -- must be a valid Workbook object')

        definition = safely(
            lambda: Item._dict_from_item_output(Item._get_item_output(session, item_id)),
            action_description=f'get properties of worksheet {item_id}',
            status=Status(errors='catalog'))
        if definition is None:
            if status.errors == 'catalog':
                return
            else:
                raise SPyRuntimeError(f'Worksheet with ID "{item_id}" does not exist')
        worksheet = Worksheet._instantiate(workbook, definition, provenance=Item.PULL)
        worksheet.pull_worksheet(session, item_id, extra_workstep_tuples=extra_workstep_tuples,
                                 include_annotations=include_annotations, include_images=include_images,
                                 include_archived=include_archived, status=status)
        return worksheet

    def pull_worksheet(self, session: Session, worksheet_id, extra_workstep_tuples=None, include_images=True,
                       status=None, include_annotations=True, include_archived: Optional[bool] = None):
        status = Status.validate(status, session)

        if include_images and not include_annotations:
            raise SPyValueError('Must specify include_images=False if you specify include_annotations=False')

        if include_annotations:
            self._annotation.pull(session, include_images=include_images,
                                  include_archived=include_archived, status=status)

        link_url = _url.SeeqURL.parse(session.public_url)
        link_url.route = _url.Route.WORKBOOK_EDIT
        link_url.folder_id = self.workbook['Ancestors'][-1] if len(self.workbook['Ancestors']) > 0 else None
        link_url.workbook_id = self.workbook.id
        link_url.worksheet_id = self.id
        self['URL'] = link_url.url

    def pull_rendered_content(self, session: Session, status: Status):
        pass

    def item_map_worksteps_key(self):
        return f'Worksteps for {self.id}'

    def push(self, context: WorkbookPushContext, pushed_workbook_id, item_map, datasource_output,
             existing_worksheet_identifiers, include_inventory, label=None):
        self._item_map_for_push = item_map

        existing_worksheet_id = None

        # After Integrated Security was introduced, we can no longer search for Worksheets using Data ID,
        # so we use the passed-in dictionary that the Workbook assembled to find existing worksheets.
        # So we call this "SPy ID" even though it's just the Data ID we would have used if searching was still
        # supported. Why not still use Data ID? Because with the introduction of templates, it's much easier to
        # choose a label that causes a conflicting Data ID in some other workbook, and it's really hard to make the
        # error (which results from a Postgres unique index constraint) and resolution guidance easy to understand.
        spy_id_with_label_suffix_maybe = self._construct_data_id(label)
        spy_id_without_label_suffix = _common.removesuffix(spy_id_with_label_suffix_maybe, f' {label}')

        if self.id in existing_worksheet_identifiers:
            existing_worksheet_id = existing_worksheet_identifiers[self.id]
        elif spy_id_with_label_suffix_maybe in existing_worksheet_identifiers:
            existing_worksheet_id = existing_worksheet_identifiers[spy_id_with_label_suffix_maybe]
        elif spy_id_without_label_suffix in existing_worksheet_identifiers:
            existing_worksheet_id = existing_worksheet_identifiers[spy_id_without_label_suffix]
        elif self.provenance == Item.CONSTRUCTOR and self.name in existing_worksheet_identifiers:
            existing_worksheet_id = existing_worksheet_identifiers[self.name]

        workbooks_api = WorkbooksApi(context.session.client)
        items_api = ItemsApi(context.session.client)
        props = list()
        if not existing_worksheet_id:
            worksheet_input = WorksheetInputV1()
            worksheet_input.name = self.definition['Name']
            worksheet_output = workbooks_api.create_worksheet(
                workbook_id=pushed_workbook_id, body=worksheet_input)  # type: WorksheetOutputV1

            if context.session.options.wants_compatibility_with(194):
                spy_id = spy_id_with_label_suffix_maybe
            else:
                spy_id = spy_id_without_label_suffix

            props.append(ScalarPropertyV1(name='SPy ID', value=spy_id))
        else:
            worksheet_output = workbooks_api.get_worksheet(workbook_id=pushed_workbook_id,
                                                           worksheet_id=existing_worksheet_id)
            props.append(ScalarPropertyV1(name='Name', value=self.definition['Name']))

        item_map[self.id] = worksheet_output.id

        props.append(ScalarPropertyV1(name='Archived', value=_common.get(self, 'Archived', False)))

        items_api.set_properties(id=worksheet_output.id, body=props)

        return worksheet_output

    @property
    def referenced_items(self):
        referenced_items = set()
        if hasattr(self, 'worksteps'):
            for workstep_id, workstep in self.worksteps.items():  # type: Workstep
                referenced_items.update(workstep.referenced_items)

        referenced_items.update(self._annotation.referenced_items)

        return list(referenced_items)

    def referenced_items_df(self):
        unique_ids = {r.id for r in self.referenced_items}
        return pd.DataFrame([self.workbook.item_inventory[_id].definition_dict for _id in unique_ids])

    @property
    def referenced_worksteps(self):
        return self._annotation.referenced_worksteps

    def find_workbook_links(self, session: Session, status: Status):
        return self._annotation.find_workbook_links(session, status)

    def find_unresolved_worksteps(self):
        # Unresolved Report workstep dependencies will be found when pushing the associated document
        dependencies_not_found = set()
        for workstep_tuple in self._annotation.find_workstep_references(item_map=self._item_map_for_push):
            referenced_workbook_id, referenced_worksheet_id, referenced_workstep_id = workstep_tuple
            if referenced_workstep_id not in self._item_map_for_push:
                dependencies_not_found.add(
                    'Workbook %s, Worksheet %s, Workstep %s' % (
                        referenced_workbook_id, referenced_worksheet_id, referenced_workstep_id))
        return dependencies_not_found

    def current_workstep(self) -> AnalysisWorkstep:
        return self.worksteps[self.definition['Current Workstep ID']]

    @staticmethod
    def _get_worksheet_json_file(workbook_folder, worksheet_id):
        return os.path.join(workbook_folder, 'Worksheet_%s.json' % worksheet_id)

    def save(self, workbook_folder, *, include_rendered_content=False, pretty_print_html=False):
        worksheet_json_file = Worksheet._get_worksheet_json_file(workbook_folder, self.id)

        with util.safe_open(worksheet_json_file, 'w', encoding='utf-8') as f:
            json.dump(self._definition, f, indent=4)

        self._annotation.save(workbook_folder, include_rendered_content=include_rendered_content,
                              pretty_print_html=pretty_print_html)

    def _load(self, workbook_folder, worksheet_id):
        worksheet_json_file = Worksheet._get_worksheet_json_file(workbook_folder, worksheet_id)

        with util.safe_open(worksheet_json_file, 'r', encoding='utf-8') as f:
            self._definition = json.load(f)

        if self.workbook['Workbook Type'] == 'Analysis':
            self._annotation = Journal.load(self, workbook_folder)
        else:
            self._annotation = Report.load(self, workbook_folder)

        self.worksteps = dict()
        folder_escaped = glob.escape(workbook_folder)
        workstep_files = util.safe_glob(os.path.join(folder_escaped, 'Worksheet_%s_Workstep_*.json' % worksheet_id))
        for workstep_file in workstep_files:
            match = re.match(r'.*?Worksheet_.*?_Workstep_(.*?).json$', workstep_file)
            workstep_id = match.group(1)
            self.worksteps[workstep_id] = Workstep.load_from_workbook_folder(self, workbook_folder, workstep_id)

    @staticmethod
    def load_from_workbook_folder(workbook, workbook_folder, worksheet_id):
        worksheet = Worksheet._instantiate(workbook, provenance=Item.LOAD)
        worksheet._load(workbook_folder, worksheet_id)
        return worksheet

    @property
    def timezone(self):
        """
        The string name of the worksheet's current timezone or None if one is not set
        """
        return self._get_timezone()

    @timezone.setter
    def timezone(self, value):
        self._set_timezone(value)

    def _set_timezone(self, timezone):
        # type: (str) -> None
        """
        Set the timezone for the current worksheet.

        A list of all timezone names is available in the pytz module:
        All timezones
            pytz.all_timezones
        Those for a specific country
            pytz.country_timezones('US')
            Where the abbreviations for countries can be found from
            list(pytz.country_names.items())

        Parameters
        ----------
        timezone : str
            The name of the desired timezone
        """
        new_workstep = self._branch_current_workstep()
        new_workstep.timezone = timezone

    def _get_timezone(self) -> Optional[str]:
        """
        Get the timezone name.

        See property "timezone"

        Returns
        -------
        {str, None}
            The string name of the worksheet's current timezone or None
            if one is not set
        """
        return self.current_workstep().timezone

    def _branch_current_workstep(self):
        if self.provenance == self.CONSTRUCTOR:
            return self.current_workstep()

        if 'Current Workstep ID' in self.definition and self.definition['Current Workstep ID']:
            new_workstep = AnalysisWorkstep(self,
                                            definition={
                                                'Data': self.worksteps[self.definition['Current Workstep ID']].data})
        else:
            new_workstep = AnalysisWorkstep(self)

        self.worksteps[new_workstep['ID']] = new_workstep
        self.definition['Current Workstep ID'] = new_workstep['ID']

        return new_workstep

    def branch_current_workstep(self):
        return self._branch_current_workstep()

    def remove_workstep(self, workstep):
        if workstep.id in self.worksteps:
            del self.worksteps[workstep.id]


class AnalysisWorksheet(Worksheet):

    def __init__(self, workbook, definition=None, *, provenance=None, add_defaults=True):
        super().__init__(workbook, definition, provenance=provenance, add_defaults=add_defaults)
        self._trend_toolbar = TrendToolbar(self)
        self._condition_table_toolbar = ConditionTableToolbar(self)

    @property
    def trend_toolbar(self):
        """
        Configuration object that exposes settings for Trend View
        """
        return self._trend_toolbar

    @property
    def condition_table_toolbar(self):
        """
        Configuration object that exposes settings for Condition Tables in Table & Charts View
        """
        return self._condition_table_toolbar

    @property
    def html(self):
        """
        Get/Set the raw HTML for the journal
        """
        return self.journal.html

    @html.setter
    def html(self, value):
        self.journal.html = value

    @property
    def journal(self):
        """
        Returns the "Journal" object, which is an "Annotation" that houses the HTML and other metadata about the
        Journal attached to this worksheet.
        """
        return self._annotation

    def workstep(self, name=None, create=True):
        if not name:
            name = _common.new_placeholder_guid()

        existing_workstep = [ws for ws in self.worksteps.values() if ws.name == name]
        if len(existing_workstep) == 1:
            return existing_workstep[0]
        elif not create:
            return None

        workstep = AnalysisWorkstep(self, {'Name': name})
        workstep.set_as_current()

        return workstep

    def pull_worksheet(self, session: Session, worksheet_id, extra_workstep_tuples=None, include_images=True,
                       status=None, include_annotations=True, include_archived: Optional[bool] = None,
                       include_referenced_worksteps=True):
        super().pull_worksheet(session, worksheet_id, extra_workstep_tuples, include_images, status,
                               include_annotations, include_archived)

        workbooks_api = WorkbooksApi(session.client)
        worksheet_output = safely(
            lambda: workbooks_api.get_worksheet(workbook_id=self.workbook.id, worksheet_id=worksheet_id),
            action_description=f'get Worksheet details for {self.workbook.id}/{worksheet_id}',
            status=status)  # type: WorksheetOutputV1
        if worksheet_output is None:
            return
        current_workstep_id = worksheet_output.workstep

        workstep_tuples_to_pull = set()
        if current_workstep_id is not None:
            workstep_tuples_to_pull.add((self.workbook.id, self.id, current_workstep_id))
            self._definition['Current Workstep ID'] = current_workstep_id

        if include_referenced_worksteps:
            for workbook_id, worksheet_id, workstep_id in self.journal.referenced_worksteps:
                if isinstance(self.journal, Journal) or worksheet_id == self.id:
                    workstep_tuples_to_pull.add((workbook_id, worksheet_id, workstep_id))

        if extra_workstep_tuples:
            for workbook_id, worksheet_id, workstep_id in extra_workstep_tuples:
                if workbook_id == self.workbook.id and worksheet_id == self.id and workstep_id is not None:
                    workstep_tuples_to_pull.add((workbook_id, worksheet_id, workstep_id))
        self._pull_worksteps(session, workstep_tuples_to_pull, status)

        if not self.worksteps:
            workstep = AnalysisWorkstep(self)
            self._definition['Current Workstep ID'] = workstep['ID']

    def _pull_worksteps(self, session: Session, workstep_tuples, status: Status):
        for workstep_tuple in workstep_tuples:
            workbook_id, worksheet_id, workstep_id = workstep_tuple
            if workstep_id not in self.worksteps:
                self.workbook.update_status('Pulling worksteps', 0)
                self.worksteps[workstep_id] = Workstep.pull(workstep_tuple, worksheet=self, session=session,
                                                            status=status)
                self.workbook.update_status('Pulling worksteps', 1)

    def pull_current_workstep(self, quiet: bool = False, status: Optional[Status] = None,
                              session: Optional[Session] = None):
        # noinspection PyUnresolvedReferences
        """
        Pulls the current workstep for the given worksheet so that the Python
        object has been updated with what the user might have changed in the
        user interface.

        Parameters
        ----------
        quiet : bool, default False
            If True, suppresses progress output. Note that when status is
            provided, the quiet setting of the Status object that is passed
            in takes precedence.

        status : spy.Status, optional
            If specified, the supplied Status object will be updated as the command
            progresses. It gets filled in with the same information you would see
            in Jupyter in the blue/green/red table below your code while the
            command is executed. The table itself is accessible as a DataFrame via
            the status.df property.

        session : spy.Session, optional
            If supplied, the Session object (and its Options) will be used to
            store the login session state. This is useful to log in to different
            Seeq servers at the same time or with different credentials.

        Example
        -------
        >>> worksheet = workbook.worksheets['My Worksheet']
        >>> worksheet.pull_current_workstep()
        """
        session = Session.validate(session)
        status = Status.validate(status, session, quiet)
        self.pull_worksheet(session, self.id, include_images=False, status=status, include_annotations=False,
                            include_referenced_worksteps=False)

    def push_current_workstep(self, quiet: bool = False, status: Optional[Status] = None,
                              session: Optional[Session] = None):
        # noinspection PyUnresolvedReferences
        """
        Pushes a new workstep for the given worksheet using the current in-memory value given by current_workstep().

        Parameters
        ----------
        quiet : bool, default False
            If True, suppresses progress output. Note that when status is
            provided, the quiet setting of the Status object that is passed
            in takes precedence.

        status : spy.Status, optional
            If specified, the supplied Status object will be updated as the command
            progresses. It gets filled in with the same information you would see
            in Jupyter in the blue/green/red table below your code while the
            command is executed. The table itself is accessible as a DataFrame via
            the status.df property.

        session : spy.Session, optional
            If supplied, the Session object (and its Options) will be used to
            store the login session state. This is useful to log in to different
            Seeq servers at the same time or with different credentials.

        Example
        -------
        >>> worksheet = workbook.worksheets['My Worksheet']
        >>> worksheet.display_items = worksheet.display_items.drop(
        >>>     worksheet.display_items[worksheet.display_items['Name'] == 'Temperature'].index)
        >>> worksheet.push_current_workstep()
        """
        session = Session.validate(session)
        status = Status.validate(status, session, quiet)
        workstep = self.current_workstep()
        workstep_input = WorkstepInputV1(data=safe_json_dumps(workstep.data))
        safely(lambda: WorkbooksApi(session.client).create_workstep(workbook_id=self.workbook.id, worksheet_id=self.id,
                                                                    no_workstep_message=False, body=workstep_input),
               status=status)

    def push(self, context: WorkbookPushContext, pushed_workbook_id, item_map, datasource_output,
             existing_worksheet_identifiers, include_inventory, label=None):
        worksheet_output = super().push(context, pushed_workbook_id, item_map, datasource_output,
                                        existing_worksheet_identifiers, include_inventory,
                                        label=label)

        pushed_current_workstep_id = None
        for workstep_id, workstep in self.worksteps.items():  # type: (str, Workstep)
            self.workbook.update_status('Pushing worksteps', 0)
            # Intentionally don't send a workstep message since it will be sent below when we set the current workstep
            pushed_workstep_id = workstep.push_to_specific_worksheet(context.session, pushed_workbook_id,
                                                                     worksheet_output, item_map, include_inventory,
                                                                     no_workstep_message=True)
            self.workbook.update_status('Pushing worksteps', 1)

            # We have to store off a per-worksheet map of worksteps because of the way they can be duplicated due
            # to duplicated worksheets and copy/pasted Journal links
            if self.item_map_worksteps_key() not in item_map:
                item_map[self.item_map_worksteps_key()] = dict()
            item_map[self.item_map_worksteps_key()][workstep_id] = pushed_workstep_id

            if workstep_id == self.definition['Current Workstep ID']:
                pushed_current_workstep_id = pushed_workstep_id

        if not pushed_current_workstep_id:
            raise SPyRuntimeError("Workstep for worksheet's 'Current Workstep ID' not found")

        workbooks_api = WorkbooksApi(context.session.client)
        # We have to do this at the end otherwise the other pushed worksheets will take precedence
        safely(lambda: workbooks_api.set_current_workstep(workbook_id=pushed_workbook_id,
                                                          worksheet_id=worksheet_output.id,
                                                          workstep_id=pushed_current_workstep_id),
               action_description=f'set {pushed_workbook_id}/{worksheet_output.id}/{pushed_current_workstep_id} '
                                  f'as the current workstep',
               status=context.status)  # type: WorksheetOutputListV1

        if context.include_annotations:
            self.journal.push(context.session, pushed_workbook_id, worksheet_output.id, item_map, datasource_output,
                              context.access_control, push_images=True, label=label)

        return worksheet_output

    def refresh_from(self, new_item, item_map: ItemMap, status: Status):
        super().refresh_from(new_item, item_map, status)

        self.journal.refresh_from(new_item.journal, item_map, status)

        for workstep_id, workstep in self.worksteps.items():
            workstep['ID'] = item_map[workstep_id]
        self.worksteps = new_item.worksteps

    def save(self, workbook_folder, *, include_rendered_content=False, pretty_print_html=False):
        super().save(workbook_folder, include_rendered_content=include_rendered_content,
                     pretty_print_html=pretty_print_html)

        for workstep_id, workstep in self.worksteps.items():
            workstep.save(workbook_folder)

    @property
    def display_items(self):
        # noinspection PyIncorrectDocstring, PyUnresolvedReferences
        """
        Add items to the display pane.

        Items in the input data frame that do not have a known store will be
        skipped.

        Parameters
        ----------
        value: dict, pd.DataFrame, pd.Series
            A pandas DataFrame with the items to be added to the display. It
            must minimally have 'ID' and 'Type' columns. Display attributes
            should be in named columns as described below.

            Type Key:
            Si = Signal, Sc = Scalar, C = Condition, M = Metric, T = Table

            ================= =================================== =============
            Display Attribute Description                         Applicability
            ================= =================================== =============
            Color             A 3-part RGB hexadecimal color spec
                              starting with "#".  E.g. #CE561B    All
            Axis Show         Boolean if the axis should be
                              shown                               Si, Sc, M
            Axis Auto Scale   Boolean if the axis should auto
                              scale                               Si, Sc, M
            Axis Limits       A dict of {'min': float,
                              'max': float} to specify the axis
                              limits. Ignored if Auto Axis Scale
                              is True.                            Si, Sc, M
            Axis Group        An identifier to specify shared
                              axes                                Si, Sc, M
            Lane              The lane a trend is plotted in      Si, Sc, M
            Axis Align        Specify the side of the plot for
                              the y-axis. 'Left' (default) or
                              'Right'.                            Si, Sc, M
            Line Style        The trend line style. Options are
                              'Solid', 'Dot', 'Short Dash',
                              'Long Dash', 'Short Dash-Dot',
                              'Short Dash-Dot-Dot',
                              'Long Dash-Dot',
                              'Long Dash-Dot-Dot', 'Dash-Dot'     Si, Sc, M
            Line Width        The width of the line.              Si, Sc, M
            Samples Display   The sample display style. Options
                              are 'Line', 'Line and Sample',
                              'Samples', 'Bars'                   Si
            Selected          Boolean indicating if item should
                              be selected                         All
            Stack             Boolean indicating if bars should
                              be stacked                          T
            ================= =================================== =============

        Returns
        -------
        pd.DataFrame
            A DataFrame with all items displayed on the worksheet, including display
            settings.
        """
        return self._get_display_items()

    @display_items.setter
    def display_items(self, value):
        self._set_display_items(items_df=value)

    @property
    def view(self):
        """
        The current view for the workstep. Valid values are

        ============ =========================================
        View         Result
        ============ =========================================
        Trend        Show the time-series trend view (default)
        Scatter Plot Show the scatter plot view
        Treemap      Show the treemap view
        Table        Show the table view
        ============ =========================================
        """
        return self._get_worksheet_view()

    @view.setter
    def view(self, value):
        self._set_worksheet_view(value)

    @property
    def display_range(self):
        # noinspection PyIncorrectDocstring
        """
        Set the display range on the worksheet

        Parameters
        ----------
        value: pandas.DataFrame, pandas.Series, dict
            The display range as a single row DataFrame, Series or dict with
            columns of 'Start' and 'End' containing the datetime objects or
            text parsable using pandas.to_datetime().

        Returns
        -------
         dict
            A dict with keys of 'Start' and 'End' and values of the ISO8601
            timestamps for the start and end of the display range.
        """
        return self._get_display_range()

    @display_range.setter
    def display_range(self, value):
        self._set_display_range(value)

    @property
    def investigate_range(self):
        # noinspection PyIncorrectDocstring
        """
        Set the investigate range on the worksheet

        Parameters
        ----------
        value: pandas.DataFrame, pandas.Series, dict
            The investigate range as a single row DataFrame, Series or dict
            with columns of 'Start' and 'End' containing the datetime objects
            or text parsable using pandas.to_datetime().

        Returns
        -------
        dict
            A dict with keys of 'Start' and 'End' and values of the ISO8601
            timestamps for the start and end of the investigate range.
        """
        return self._get_investigate_range()

    @investigate_range.setter
    def investigate_range(self, value):
        self._set_investigate_range(value)

    @property
    @deprecated(reason='Use self.table_date_display instead')
    def scorecard_date_display(self):
        return self.table_date_display

    @scorecard_date_display.setter
    @deprecated(reason='Use self.table_date_display instead')
    def scorecard_date_display(self, value):
        self.table_date_display = value

    @property
    @deprecated(reason='Use self.table_date_format instead')
    def scorecard_date_format(self):
        return self.table_date_format

    @scorecard_date_format.setter
    @deprecated(reason='Use self.table_date_format instead')
    def scorecard_date_format(self, value):
        self.table_date_format = value

    @property
    def table_date_display(self):
        """
        The dates that should be displayed for tables.
        Valid values are:

        =============== ================================
        Date Display    Result
        =============== ================================
        None            No date display
        'Start'         Start of the time period only
        'End'           End of the time period only
        'Start And End' Start and end of the time period
        =============== ================================
        """
        return self._get_table_date_display()

    @table_date_display.setter
    def table_date_display(self, value):
        self._set_table_date_display(value)

    @property
    def table_date_format(self):
        """
        The string defining the date format. Formats are parsed using
        momentjs. The full documentation for the momentjs date parsing
        can be found at https://momentjs.com/docs/#/displaying/

        Examples
        --------
        "d/m/yyy" omitting leading zeros (eg, 4/27/2020): l
        "Mmm dd, yyyy, H:MM AM/PM" (eg, Apr 27, 2020 5:00 PM) : lll
        "H:MM AM/PM" (eg, "5:00 PM"): LT
        """
        return self._get_table_date_format()

    @table_date_format.setter
    def table_date_format(self, value):
        self._set_table_date_format(value)

    @property
    def table_mode(self):
        """
        The string defining the Table view mode. Can be either "simple"
        or "condition".
        """
        return self._get_table_mode()

    @table_mode.setter
    def table_mode(self, value):
        self._set_table_mode(value)

    @property
    def scatter_plot_series(self):
        # noinspection PyUnresolvedReferences, PyIncorrectDocstring
        """
        Get/Set the IDs of the item plotted on the x-axis of a scatter plot

        Parameters
        ----------
        value : dict or None
            A dict with keys of the axis name (either 'X' or 'Y') and values of
            dicts, Series, or one-row DataFrames with a column of 'ID' for the
            item to use for the axis.

        Returns
        -------
        dict or None
            A dict with keys of 'X' and 'Y' and values of dicts with a key of
            'ID' and a value of either the Seeq ID of the item or None if not
            specified. Returns None if neither is specified

        Example
        -------
        >>> series = worksheet.scatter_plot_series
        >>> print(series)
            {
                'X': {'ID': '57ADF442-C571-4B19-B358-E03D77CE68B4'},
                'Y': {'ID': '3E4571C8-66C0-428E-8B04-FE0FB9F140BB'}
            }
        """
        return self._get_scatter_plot_series()

    @scatter_plot_series.setter
    def scatter_plot_series(self, value):
        self._set_scatter_plot_series(value)

    def _get_display_range(self) -> Optional[dict]:
        """
        Get the display range for the workstep

        See worksheet property "display_range" for docs

        :return:
        """
        return self.current_workstep().display_range

    def _set_display_range(self, display_range: dict):
        new_workstep = self._branch_current_workstep()
        new_workstep.display_range = display_range

    def _get_investigate_range(self) -> Optional[dict]:
        """
        Get the investigate range of the current workstep

        See worksheet property "investigate_range" for docs
        """
        return self.current_workstep().investigate_range

    def _set_investigate_range(self, investigate_range: dict):
        new_workstep = self._branch_current_workstep()
        new_workstep.investigate_range = investigate_range

    def _get_display_items(self) -> Optional[pd.DataFrame]:
        """
        Get the items of a given type displayed on the worksheet at the current workstep, regardless of the worksheet
        view

        Returns
        -------
        {pandas.DataFrame None}
            A list of the items present on the worksheet at the current workstep, or None if there is no current
            workstep ID or the current workstep has no data
        """
        return self.current_workstep().display_items

    def _set_display_items(self, items_df: pd.DataFrame):
        """
        Get the display items

        See worksheet property "display items" for docs
        """
        new_workstep = self._branch_current_workstep()
        new_workstep.display_items = items_df

    def _set_worksheet_view(self, view_key='Trend'):
        """
        Set the view key

        See worksheet property "view"
        """
        new_workstep = self._branch_current_workstep()
        new_workstep.view = view_key

    def _get_worksheet_view(self):
        return self.current_workstep().view

    def _set_table_date_display(self, date_display):
        """
        Set the date display

        See worksheet property "table_date_display" for docs
        :param date_display:
        :return:
        """
        new_workstep = self._branch_current_workstep()
        new_workstep.table_date_display = date_display

    def _get_table_date_display(self):
        return self.current_workstep().table_date_display

    def _set_table_date_format(self, date_format):
        """
        Set the table date format

        See worksheet property "table_date_format" for docs
        :param date_format:
        :return:
        """
        new_workstep = self._branch_current_workstep()
        new_workstep.table_date_format = date_format

    def _get_table_date_format(self):
        return self.current_workstep().table_date_format

    def _set_table_mode(self, mode):
        """
        Set the Table view mode

        See worksheet property "table_mode" for docs
        """
        new_workstep = self._branch_current_workstep()
        new_workstep.table_mode = mode

    def _get_table_mode(self):
        return self.current_workstep().table_mode

    def _set_scatter_plot_series(self, series_ids):
        """
        Set the IDs of the series to use for the axes of scatter plots

        Parameters
        ----------
        series_ids : dict
            A dict with keys of the axis name (either 'X' or 'Y') and values of
            the Seeq IDs of the items.
        """
        new_workstep = self._branch_current_workstep()
        new_workstep.scatter_plot_series = series_ids

    def _get_scatter_plot_series(self):
        return self.current_workstep().scatter_plot_series

    @staticmethod
    def add_display_columns(df, inplace=True):
        """
        Add the display attribute columns to a pandas.DataFrame

        Parameters
        ----------
        df : pandas.DataFrame
            The DataFrame that will have the columns added

        inplace : boolean, default True
            If True, the columns of the DataFrame passed in will be modified and
            None will be returned. If False a copy of df with the new columns
            will be returned

        Returns
        -------
        {None, pandas.DataFrame}
            None if inplace == True. A pandas.DataFrame deep copy with the
            display attribute columns added if inplace == False
        """
        return AnalysisWorkstep.add_display_columns(df, inplace)


class TopicDocument(Worksheet):

    @property
    def report(self):
        """
        Returns the "Report" object, which is an "Annotation" that houses the HTML and other metadata about the
        TopicDocument. This object also includes dictionaries that contain the Content, DateRange and AssetSelection
        objects.
        """
        return self._annotation

    def pull_rendered_content(self, session: Session, status: Status):
        status = Status.validate(status, session)

        if self.report:
            self.report.pull_rendered_content(session, status)

    def push(self, context: WorkbookPushContext, pushed_workbook_id, item_map, datasource_output,
             existing_worksheet_identifiers, include_inventory, label=None):
        worksheet_output = super().push(context, pushed_workbook_id, item_map, datasource_output,
                                        existing_worksheet_identifiers, include_inventory,
                                        label=label)

        if context.include_annotations:
            self.report.push(context.session, pushed_workbook_id, worksheet_output.id, item_map, datasource_output,
                             context.access_control, push_images=True, label=label)

        return worksheet_output

    def refresh_from(self, new_item, item_map: ItemMap, status: Status):
        super().refresh_from(new_item, item_map, status)

        self.report.refresh_from(new_item.report, item_map, status)

    def render_template(self, text=None, filename=None, **kwargs):
        try:
            from mako.lookup import TemplateLookup
            from mako.template import Template
        except ImportError:
            raise SPyDependencyNotFound('`Mako` is required to render templates. Please install it using '
                                        f'`pip install seeq-spy[templates]` to use this feature.')
        if filename:
            sibling_folder_lookup = TemplateLookup(directories=[os.path.dirname(filename)])
            template = Template(filename=filename, lookup=sibling_folder_lookup, input_encoding='utf-8')
        elif text:
            template = Template(text=text)
        else:
            raise SPyValueError('Either html or filename must be supplied')

        def _display(display, date_range=None, *, size='medium', shape='rectangle', width=None, height=None,
                     scale=1.0, selector='', asset_selection=None, summary_type=None, summary_value=None):
            if not isinstance(display, AnalysisWorkstep):
                raise SPyValueError(
                    f'The first argument you gave display() is "{display}", of type "{type(date_range)}" - '
                    f'check your @Asset.Display function. It needs to be decorated with @Asset.Display '
                    f'and return a valid AnalysisWorkstep.')

            if not isinstance(date_range, DateRange) and date_range is not None:
                raise SPyValueError(f'The date_range specified is "{date_range}", of type "{type(date_range)}" - check '
                                    f'the date range definition in your Asset class function, it needs to be decorated '
                                    f'with @Asset.DateRange and return a valid date range')

            if not isinstance(asset_selection, AssetSelection) and asset_selection is not None:
                raise SPyValueError(f'The asset_selection specified is "{asset_selection}", of type "'
                                    f'{type(asset_selection)}" - check the asset selection definition in your Asset '
                                    f'class function, it needs to be decorated with @Asset.AssetSelection and return '
                                    f'a valid asset selection')

            return self.report.get_embedded_content_html(display, date_range, size=size, shape=shape, width=width,
                                                         height=height, scale=scale, selector=selector,
                                                         asset_selection=asset_selection,
                                                         summary_type=summary_type, summary_value=summary_value)

        def _plot(plot_function, date_range):
            return self.report.add_plot_to_render(plot_function, date_range)

        folder_for_relative_paths = os.path.dirname(filename)

        # noinspection PyShadowingNames
        def _image(filename=None, *, buffer=None, image_format=None):
            if not os.path.isabs(filename):
                filename = os.path.join(folder_for_relative_paths, filename)
            return self.report.add_image(filename=filename, buffer=buffer, image_format=image_format, just_src=True)

        kwargs['display'] = _display
        kwargs['plot'] = _plot
        kwargs['image'] = _image

        self.html = template.render(**kwargs)

    @property
    def html(self):
        """
        Get/Set the raw HTML for the topic document
        """
        return self.report.html

    @html.setter
    def html(self, value):
        self.report.html = value

    @property
    def date_ranges(self):
        return self.report.date_ranges

    @property
    def content(self):
        return self.report.content

    @property
    def asset_selections(self):
        return self.report.asset_selections

    @property
    def schedule(self):
        # noinspection PyUnresolvedReferences, HttpUrlsUsage, PyIncorrectDocstring
        r"""
        Get/Set the schedule for the topic document. The schedule determines whether a document is a live,
        auto-updating document or a scheduled document.

        Parameters
        ----------
        value : dict or None
            A dictionary of a 'Background' and a 'Cron Schedule' for document's schedule

                ===================== =============================================
                Input Column          Schedule Attribute
                ===================== =============================================
                Background            Whether the document is scheduled or live
                Cron Schedule         The document's update schedule as a cron
                                      schedule. This should be a list of cron
                                      expressions such as ['0 \*/5 \* \* \* ?']. See
                                      http://www.quartz-scheduler.org/documentation/quartz-2.3.0/tutorials/crontrigger.html
                                      for more information
                ===================== =============================================

        Returns
        -------
        dict
            The current dictionary containing the 'Background' variable and 'Cron Schedule' for the document's schedule
        """
        return self.report.schedule

    @schedule.setter
    def schedule(self, value):
        self.report.schedule = value


@deprecated(reason='Use spy.utils.pull_worksheet_via_url instead')
def get_analysis_worksheet_from_url(url: str, include_archived: bool = False, quiet: Optional[bool] = False,
                                    session: Optional[Session] = None):
    return pull_worksheet_via_url(url=url, include_archived=include_archived, quiet=quiet, session=session,
                                  minimal=False)


def pull_worksheet_via_url(url: str, *, minimal: bool = True, include_archived: bool = False,
                           quiet: Optional[bool] = False, session: Optional[Session] = None):
    """
    Get a worksheet from a valid Workbench Analysis URL.

    Parameters
    ----------
    url : str
        The URL of the Workbench Analysis
    minimal : bool, default True
        If True, only the minimal amount of information will be pulled from the server. This is useful if you are
        only interested in, for example, knowing what display_items are present.
    include_archived : bool, default False
        If True, returns the worksheet even if the workbook is archived. If False, raises an exception
        if the workbook is archived.
    quiet : bool, default False
        If True, suppresses progress output.
    session : spy.Session, optional
        If supplied, the Session object used for the connection to Seeq
        Server. If not supplied, the default session is used.

    Returns
    -------
    {Worksheet}
        The Worksheet object
    """
    workbook_id = _url.get_workbook_id_from_url(url)
    worksheet_id = _url.get_worksheet_id_from_url(url)

    if workbook_id is None or worksheet_id is None:
        raise SPyValueError(f'The supplied URL is not a valid Seeq address. Verify that both the workbook ID and '
                            f'worksheet ID are valid Seeq references')

    session = Session.validate(session)
    status = Status(errors='catalog', quiet=quiet)
    _login.validate_login(session, status)

    workbook_search = spy.workbooks.search({'ID': workbook_id}, status=status, session=session)
    if len(workbook_search) == 0:
        raise SPyRuntimeError(f'Could not find workbook with ID "{workbook_id}"')
    workbooks = spy.workbooks.pull(workbook_search, include_referenced_workbooks=False, include_inventory=False,
                                   include_annotations=not minimal, include_images=not minimal,
                                   include_access_control=not minimal, specific_worksheet_ids=[worksheet_id],
                                   status=Status(quiet=quiet), session=session)
    if not workbooks:
        raise SPyValueError(f"An error occurred while pulling Worksheet '{worksheet_id}' from Workbook '{workbook_id}'")
    workbook = workbooks[0]

    if _common.get(workbook.definition, 'Archived') and not include_archived:
        raise SPyValueError(f"Workbook '{workbook_id}' is archived. Supply 'include_archived=True' if "
                            f"you want to retrieve the items of an archived workbook")

    if workbook.definition['Workbook Type'] != 'Analysis':
        raise SPyValueError(f'URL must be for a valid Workbench Analysis. '
                            f'You supplied a URL for a {workbook.definition["Workbook Type"]}.')

    try:
        worksheet = [x for x in workbook.worksheets if x.id == worksheet_id][0]
        if _common.get(worksheet.definition, 'Archived') and not include_archived:
            raise SPyValueError(f"Worksheet '{worksheet_id}' is archived. Supply 'include_archived=True' if "
                                f"you want to retrieve archived items")
        return worksheet
    except IndexError:
        raise SPyRuntimeError(f'Worksheet with ID "{worksheet_id}" does not exist in workbook "{workbook.name}"')


class WorksheetList(ItemList):
    def __init__(self, workbook: spy.workbooks.Workbook, seq=()):
        super().__init__(seq)
        self.workbook = workbook
        self._validate_and_transfer_if_necessary()

    def _validate_and_transfer_if_necessary(self):
        worksheets = dict()
        for worksheet in self:
            if worksheet.id in worksheets:
                raise SPyValueError(f'Worksheets with duplicate IDs exist within this workbook:\n'
                                    f'{worksheet}\n{worksheets[worksheet.id]}')

            worksheets[worksheet.id] = worksheet
            if worksheet.workbook is not self.workbook:
                # When a worksheet transfers over from a different workbook, we need to bring the workbook's item
                # inventory and datasource maps over as well.
                self.workbook.item_inventory.update(worksheet.workbook.item_inventory)
                self.workbook.datasource_maps.extend(worksheet.workbook.datasource_maps)

                worksheet.workbook = self.workbook

    # noinspection PyTypeChecker
    def __getitem__(self, key) -> Worksheet:
        return super().__getitem__(key)

    def __setitem__(self, key, val: Worksheet):
        retval = super().__setitem__(key, val)
        self._validate_and_transfer_if_necessary()
        return retval

    def append(self, __object: Worksheet) -> None:
        super().append(__object)
        self._validate_and_transfer_if_necessary()

    def extend(self, __iterable: Iterable[Worksheet]) -> None:
        super().extend(__iterable)
        self._validate_and_transfer_if_necessary()



from dataclasses import dataclass
from typing import Any, List, Optional

from seeq.spy._session import Session
from seeq.spy._status import Status


@dataclass
class WorkbookPushContext:
    access_control: Optional[str]
    datasource: Optional[str]
    dummy_items_workbook_context: Optional[Any]
    include_annotations: Optional[bool]
    override_max_interp: Optional[bool]
    owner: Optional[str]
    reconcile_inventory_by: str
    global_inventory: str
    session: Session
    specific_worksheet_ids: Optional[List[str]]
    status: Status



from seeq.spy._dependencies import Dependencies
from seeq.spy._errors import SPyDependencyNotFound
from seeq.spy.widgets._ipy_utils import WidgetLogHandler
from seeq.spy.widgets._widgets import SeeqItemSelect, LogWindowWidget


class DataLabEnvMgr:
    def __new__(cls, *args, **kwargs):
        raise SPyDependencyNotFound(f'`seeq-data-lab-env-mgr` is not installed. Please use `pip install seeq-spy[widgets]` '
                                    f'to use this feature.')


try:
    from seeq.data_lab_env_mgr import DataLabEnvMgr
except ImportError:
    pass

__all__ = ['LogWindowWidget', 'SeeqItemSelect', 'WidgetLogHandler', 'DataLabEnvMgr']



import json
import re

from seeq.spy import _common
from seeq.spy._errors import SPyDependencyNotFound


class MustachioedAnnotation:
    """
    Wraps a piece of HTML and exposes any Mustache variables in a way that is useful to the templating system. Also
    facilitates rendering of a template with Mustache variables replaced.
    """

    INVALID_MUSTACHE_CHARS = r'\.'
    INVALID_MUSTACHE_CHARS_REGEX = re.compile('[' + INVALID_MUSTACHE_CHARS + ']')

    _html: str
    _code_dict: dict

    def __init__(self, html, resolve_content_key_func=None):
        self._html = _common.fix_up_ckeditor_curly_brace_weirdness(html)

        if resolve_content_key_func is not None:
            self._html = MustachioedAnnotation._parameterize_seeq_content_ids(self._html, resolve_content_key_func)

        self._html = MustachioedAnnotation.fix_up_images(self._html)

        self._validate()

    @property
    def html(self):
        return self._html

    @property
    def code(self):
        return json.dumps(self._code_dict, indent=4).replace(': null', ': None')

    @property
    def code_dict(self):
        return self._code_dict

    @staticmethod
    def _parameterize_seeq_content_ids(html, resolve_content_key_func):
        return re.sub(r'data-seeq-content="([^"]+)"',
                      lambda m: f'data-seeq-content="{{{{{resolve_content_key_func(m.group(1))}}}}}"',
                      html)

    @staticmethod
    def fix_up_images(html):
        replacements = list()
        for img_tag_match in re.finditer(r'<img [^>]+>', html):
            # The reason for the \W? is that CKEditor puts in a zero-width space (Unicode character u200b) after the
            # first curly brace
            alt_attr_regex = r' alt="\{\W?\{([^}]+)}\W?}"'
            img_tag = img_tag_match.group(0)

            alt_attr_match = re.search(alt_attr_regex, img_tag)
            if alt_attr_match is None:
                continue

            var_name = alt_attr_match.group(1)
            new_img_tag = re.sub(r' src=\"/api/annotations/([^\"]+)\"', f' src="{{{{[Image] {var_name}}}}}"', img_tag)
            new_img_tag = re.sub(alt_attr_regex, f' alt="{{{{[AltText] {var_name}}}}}"', new_img_tag)
            replacements.append((img_tag, new_img_tag))

        for old, new in replacements:
            html = html.replace(old, new)
        return html

    # noinspection PyBroadException
    @staticmethod
    def invalid_mustache_characters():
        try:
            import chevron
        except ImportError:
            raise SPyDependencyNotFound(
                f'`Chevron` is required to use this feature. Please use '
                f'`pip install seeq-spy[templates]` to use this feature.')
        invalid_chars = str()
        for i in range(0, 65536):

            try:
                output = chevron.render('{{z' + chr(i) + 'z}}', {'z' + chr(i) + 'z': 'Worked!'})
                if output != 'Worked!':
                    invalid_chars += chr(i)
            except Exception:
                invalid_chars += chr(i)

        return invalid_chars

    @staticmethod
    def _sanitize_mustache_token(token):
        return MustachioedAnnotation.INVALID_MUSTACHE_CHARS_REGEX.sub('-', token)

    def _validate(self):
        matches = re.finditer(r'({{{?)([!#^/>=&]?)([^}]+)(}?}})', self.html)
        replacements = list()
        root_dict = dict()
        stack = list([root_dict])
        current_dict = root_dict
        for match in matches:
            prefix = match.group(1)
            qualifier = match.group(2)
            name = match.group(3)
            suffix = match.group(4)

            if name is None:
                continue

            sanitized_name = MustachioedAnnotation._sanitize_mustache_token(name)

            if qualifier is not None:
                replacements.append((match.group(0), prefix + qualifier + sanitized_name + suffix))
            else:
                replacements.append((match.group(0), prefix + sanitized_name + suffix))

            new_list = [dict()]
            if qualifier in ['#', '^']:
                stack.append(current_dict)
                if sanitized_name in current_dict:
                    current_dict = current_dict[sanitized_name][0]
                else:
                    current_dict[sanitized_name] = new_list
                    current_dict = new_list[0]
            elif qualifier == '/':
                current_dict = stack.pop()
            else:
                current_dict[sanitized_name] = None

        for _old, _new in replacements:
            self._html = self._html.replace(_old, _new)

        self._code_dict = root_dict

    def render(self, parameters: dict) -> str:
        try:
            import chevron
        except ImportError:
            raise SPyDependencyNotFound(
                f'`Chevron` is required to use this feature. Please use '
                f'`pip install seeq-spy[templates]` to use this feature.')

        data = dict()
        if parameters is not None:
            for key, value in parameters.items():
                data[key] = ('{{' + key + '}}') if value is None else value

        return chevron.render(self.html, data)



from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import Optional

import pandas as pd

from seeq import spy
from seeq.spy import _common
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy._status import Status


@dataclass
class WorkbookContext:
    workbook_object: Optional[spy.workbooks.Analysis] = None
    worksheet_object: Optional[spy.workbooks.AnalysisWorksheet] = None
    folder_id: Optional[str] = None
    to_push: list = field(default_factory=list)
    item_map: spy.workbooks.ItemMap = None

    def _lookup(self, _id):
        if self.item_map is not None and _id in self.item_map:
            return self.item_map[_id]
        else:
            return _id

    @property
    def workbook_id(self):
        if self.workbook_object is None:
            return None

        return self._lookup(self.workbook_object.id)

    @property
    def worksheet_id(self):
        if self.worksheet_object is None:
            return None

        return self._lookup(self.worksheet_object.id)

    def determine_folder_id(self):
        # This is called deliberately at the top of the push function so that it's calculated once and then used
        # consistently throughout, even if the workbook_object changes.

        if self.workbook_object is None:
            return None

        self.folder_id = None
        if not isinstance(self.workbook_object, spy.workbooks.AnalysisTemplate):
            if (_common.present(self.workbook_object, 'Ancestors') and
                    len(self.workbook_object['Ancestors']) > 0):
                self.folder_id = self.workbook_object['Ancestors'][-1]

        # Here we cover the case where an admin is pushing data to a single workbook that they don't own. In R51,
        # there's a proper user folder in the path that can be used as the folder_id. In R50 there is no such user
        # folder, so the folder_id will end up being __All__ or __Shared__, which is not viable to use as a parent
        # folder. But in the case where we're pushing to only the primary workbook, then we don't need to supply the
        # folder_id because the workbook is not being moved and we're not adding other sibling workbooks (and therefore
        # we don't need to know the actual folder that the workbook is in).
        if (self.folder_id in spy.workbooks.SYNTHETIC_FOLDERS and len(self.to_push) == 1 and
                self.to_push[0] == self.workbook_object):
            self.folder_id = None

    def get_workbook_url(self, session: Session):
        if self.workbook_object is None:
            return None

        return '%s/%sworkbook/%s/worksheet/%s' % (
            session.public_url,
            (self.folder_id + '/') if self.folder_id is not None else '',
            self.workbook_id,
            self.worksheet_id
        )

    @staticmethod
    def from_id(session: Session, status: Status, workbook_arg, worksheet_arg) -> \
            WorkbookContext:
        primary_workbooks = spy.workbooks.pull(pd.DataFrame([{
            'ID': _common.sanitize_guid(workbook_arg),
            'Type': 'Workbook',
            'Workbook Type': 'Analysis'
        }]),
            include_inventory=False,
            # Pulling worksheets can be expensive. If we just want to push items without changing the UI, skip that.
            specific_worksheet_ids=list() if worksheet_arg is None else None,
            status=status.create_inner('Pull Workbook', quiet=True),
            session=session)

        if len(primary_workbooks) == 0 or primary_workbooks[0].__class__.__name__ == 'Workbook':
            raise SPyRuntimeError(f'Workbook with ID "{_common.sanitize_guid(workbook_arg)}" not found')

        # noinspection PyTypeChecker
        primary_workbook: spy.workbooks.Analysis = primary_workbooks[0]
        if isinstance(worksheet_arg, str):
            primary_worksheet = primary_workbook.worksheet(worksheet_arg)
        elif isinstance(workbook_arg, spy.workbooks.AnalysisWorksheetTemplate):
            primary_worksheet = workbook_arg
        else:
            primary_worksheet = None

        return WorkbookContext(workbook_object=primary_workbook, worksheet_object=primary_worksheet)

    @staticmethod
    def from_list(session: Session, status: Status, workbook_arg, datasource_arg):
        analysis_workbooks = [w for w in workbook_arg if w['Workbook Type'] == 'Analysis']
        if len(analysis_workbooks) == 0:
            raise SPyValueError(
                f'workbook argument is a WorkbookList that contains no Analysis workbooks. You must supply at '
                f'least one (and ideally only one, otherwise things can get confusing) Analysis workbook in the '
                f'list')

        primary_workbook = analysis_workbooks[0]

        if len(analysis_workbooks) > 1:
            status.warn(f'workbook argument is a WorkbookList that contains more than one Analysis workbook. As a '
                        f'result, the first workbook "{primary_workbook}" is the one to which '
                        f'signals/conditions/scalars (etc) will be pushed')

        push_workbooks_df = spy.workbooks.push(workbook_arg, datasource=datasource_arg, specific_worksheet_ids=list(),
                                               include_inventory=False,
                                               status=status.create_inner('Create Workbook', quiet=True),
                                               session=session)

        return WorkbookContext(workbook_object=primary_workbook, to_push=workbook_arg,
                               item_map=push_workbooks_df.spy.item_map)

    @staticmethod
    def from_analysis_template(session: Session, status: Status, workbook_arg, datasource_arg):
        push_workbooks_df = spy.workbooks.push(workbook_arg, datasource=datasource_arg, specific_worksheet_ids=list(),
                                               include_inventory=False,
                                               status=status.create_inner('Create Workbook', quiet=True),
                                               session=session)

        # noinspection PyTypeChecker
        primary_workbook: spy.workbooks.Analysis = push_workbooks_df.spy.output[0]

        return WorkbookContext(workbook_object=primary_workbook, to_push=[workbook_arg])

    @staticmethod
    def from_string(session: Session, status: Status, workbook_arg, worksheet_arg,
                    datasource_arg):
        search_query, workbook_name = WorkbookContext.create_analysis_search_query(workbook_arg)
        search_df = spy.workbooks.search(search_query, quiet=True, session=session)
        workbooks_to_push = list()
        worksheet_object = None
        if len(search_df) == 0:
            workbook_object = spy.workbooks.Analysis({'Name': workbook_name})
            path = _common.get(search_query, 'Path',
                               # If the user doesn't specify a path, default to directly within the user's home folder
                               default=_common.PATH_ROOT)
            inner_status = status.create_inner('Create Workbook', quiet=True)
            if isinstance(worksheet_arg, spy.workbooks.AnalysisWorksheetTemplate):
                workbooks_to_push = [workbook_object]
                workbook_object.worksheets.append(worksheet_arg)

                # Don't push any worksheets, since we can't resolve template parameters yet
                specific_worksheet_ids = list()

                spy.workbooks.push(
                    workbook_object, path=path, include_inventory=False, datasource=datasource_arg,
                    specific_worksheet_ids=specific_worksheet_ids, status=inner_status, session=session)
            else:
                worksheet_name = worksheet_arg if isinstance(worksheet_arg, str) else _common.DEFAULT_WORKSHEET_NAME
                worksheet_object = workbook_object.worksheet(worksheet_name)

                spy.workbooks.push(workbook_object, path=path, include_inventory=False, datasource=datasource_arg,
                                   status=inner_status, session=session)
        else:
            # noinspection PyTypeChecker
            workbook_object: spy.workbooks.Analysis = spy.workbooks.pull(
                search_df,
                include_inventory=False,
                # Pulling worksheets can be expensive. If we just want to push items without changing the UI, skip that.
                specific_worksheet_ids=list() if worksheet_arg is None else None,
                status=status.create_inner('Pull Workbook', quiet=True),
                session=session)[0]

            if isinstance(worksheet_arg, spy.workbooks.AnalysisWorksheetTemplate):
                workbooks_to_push = [workbook_object]
                worksheet_object = worksheet_arg
                if worksheet_object.name in workbook_object.worksheets:
                    existing_worksheet = workbook_object.worksheets[worksheet_object.name]
                    worksheet_object['ID'] = existing_worksheet['ID']
                    workbook_object.worksheets[worksheet_object['ID']] = worksheet_object
                else:
                    workbook_object.worksheets.append(worksheet_object)

            elif worksheet_arg is not None:
                worksheet_name = worksheet_arg if isinstance(worksheet_arg, str) else _common.DEFAULT_WORKSHEET_NAME
                worksheet_object = workbook_object.worksheet(worksheet_name)

        return WorkbookContext(workbook_object=workbook_object, worksheet_object=worksheet_object,
                               to_push=workbooks_to_push)

    @staticmethod
    def from_args(session: Session, status: Status, workbook_arg, worksheet_arg, datasource_arg) -> WorkbookContext:
        if _common.is_guid(workbook_arg):
            workbook_context = WorkbookContext.from_id(session, status, workbook_arg, worksheet_arg)
        elif isinstance(workbook_arg, list):
            workbook_context = WorkbookContext.from_list(session, status, workbook_arg, datasource_arg)
        elif isinstance(workbook_arg, spy.workbooks.AnalysisTemplate):
            workbook_context = WorkbookContext.from_analysis_template(session, status, workbook_arg, datasource_arg)
        elif workbook_arg is not None:
            workbook_context = WorkbookContext.from_string(session, status, workbook_arg, worksheet_arg, datasource_arg)
        else:
            workbook_context = WorkbookContext()

        workbook_context.determine_folder_id()

        return workbook_context

    @staticmethod
    def create_analysis_search_query(workbook) -> tuple[dict[str, str], str]:
        workbook_spec_parts = _common.path_string_to_list(workbook)
        search_query = dict()
        if len(workbook_spec_parts) > 1:
            search_query['Path'] = _common.path_list_to_string(workbook_spec_parts[0:-1])
            workbook_name = workbook_spec_parts[-1]
        else:
            workbook_name = workbook_spec_parts[0]
        search_query['Name'] = f'/^{re.escape(workbook_name)}$/'
        search_query['Workbook Type'] = 'Analysis'
        return search_query, workbook_name



from __future__ import annotations

import copy
import datetime
import re
import types
from typing import Optional, Tuple, Dict

from seeq.sdk import *
from seeq.spy import _login

# *********************************************************************************************************************
#
# This file is a direct port of several TypeScript functions from
# client/packages/webserver/app/src/annotation/reportContent.utilities.ts
# There should be NO CHANGES IN LOGIC in this file as compared to that TypeScript file, down to the slightest detail.
# Otherwise, the frontend may not correctly process the Date Ranges that are pushed by SPy and likewise SPy may not
# process the Date Ranges that are pulled from the frontend.
#
# *********************************************************************************************************************


QUARTZ_CRON_PENDING_INPUT = '59 59 23 31 12 ? 2099'

DURATION_SCALAR_UNITS = [
    's', 'second', 'seconds', 'min', 'minute', 'minutes', 'h', 'hr', 'hrs', 'hour', 'hours', 'day', 'd', 'days',
    'week', 'wk', 'weeks', 'month', 'mo', 'months', 'year', 'y', 'yr', 'years',
    'ns', 'ms'
]

OFFSET_DIRECTION = types.SimpleNamespace(
    PAST='past',
    FUTURE='future'
)

CAPSULE_SELECTION = types.SimpleNamespace(
    STRATEGY=types.SimpleNamespace(
        CLOSEST_TO='closestTo',
        OFFSET_BY='offsetBy',
    ),
    REFERENCE=types.SimpleNamespace(
        START='start',
        END='end',
    ),
)

DEFAULT_DATE_RANGE = {
    'auto': {
        'enabled': False,
        'duration': datetime.timedelta(days=1).total_seconds() * 1000,
        'offset': {'value': 0, 'units': 'min'},
        'offsetDirection': OFFSET_DIRECTION.PAST,
        'cronSchedule': [QUARTZ_CRON_PENDING_INPUT],
        'noCapsuleFound': False,
    },
    'condition': {
        'strategy': CAPSULE_SELECTION.STRATEGY.CLOSEST_TO,
        'reference': CAPSULE_SELECTION.REFERENCE.END,
        'offset': 1,
        'range': {},
    },
    'range': {},
    'enabled': True,
}


# export const formatDateRangeFromApiOutput = (dateRangeOutput: DateRangeOutputV1): DateRange => {
def format_date_range_from_api_output(date_range_output: DateRangeOutputV1) -> dict:
    # let formulaFormValid = false;
    formula_form_valid = False
    # const dateRange: any = {};
    # _.defaultsDeep(dateRange, DEFAULT_DATE_RANGE);
    date_range: dict = copy.deepcopy(DEFAULT_DATE_RANGE)
    # _.assign(dateRange, _.pick(dateRangeOutput, ['name', 'id', 'description']));
    date_range.update({'name': date_range_output.name,
                       'id': date_range_output.id,
                       'description': date_range_output.description})
    # noinspection PyUnresolvedReferences
    # _.assign(dateRange.condition, _.pick(dateRangeOutput.condition, ['name', 'id', 'isRedacted']));
    if date_range_output.condition is not None:
        date_range['condition'].update({'name': date_range_output.condition.name,
                                        'id': date_range_output.condition.id,
                                        'isRedacted': date_range_output.condition.is_redacted})
    # noinspection PyUnresolvedReferences
    # _.assign(dateRange.auto, {
    #   enabled: dateRangeOutput.isAutoUpdating,
    # });
    date_range['auto'].update({'enabled': date_range_output.is_auto_updating})
    # dateRange.enabled = dateRangeOutput.isEnabled;
    date_range['enabled'] = date_range_output.is_enabled
    # dateRange.reportId = dateRangeOutput.report?.id;
    date_range['reportId'] = date_range_output.report.id if date_range_output.report is not None else None
    # dateRange.isArchived = dateRangeOutput.isArchived;
    date_range['isArchived'] = date_range_output.is_archived

    #
    # // The backend gives us back ISO8601 timestamps, but expects milliseconds back.
    # if (dateRangeOutput.dateRange?.start) {
    #   dateRange.range.start = moment.utc(dateRangeOutput.dateRange.start).valueOf();
    # }
    if date_range_output.date_range is not None and date_range_output.date_range.start is not None:
        date_range['range']['start'] = int(_login.parse_content_datetime_with_timezone(
            session=None, dt=date_range_output.date_range.start).value / 1_000_000)
    #
    # if (dateRangeOutput.dateRange?.end) {
    #   dateRange.range.end = moment.utc(dateRangeOutput.dateRange.end).valueOf();
    # }
    if date_range_output.date_range is not None and date_range_output.date_range.end is not None:
        date_range['range']['end'] = int(_login.parse_content_datetime_with_timezone(
            session=None, dt=date_range_output.date_range.end).value / 1_000_000)
    #
    # try {
    try:
        # // See .createDateRangeFormula() for expected formula formats
        #
        # // Fixed, no condition
        # if (!dateRangeOutput.condition?.id && dateRangeOutput.formula?.match(/^capsule\(.*\)$/)) {
        if ((date_range_output.condition is None or date_range_output.condition.id is None) and
                (date_range_output.formula is not None and re.search(r'^capsule\(.*\)$', date_range_output.formula))):
            # // Nothing additional needs to be extracted from the formula
            # formulaFormValid = true;
            formula_form_valid = True

        # }
        #
        # const setConditionProperties = (
        #   searchStart?: string,
        #   searchEnd?: string,
        #   maxDuration?: string,
        #   columns?: string,
        #   sortBy?: string,
        #   sortAsc?: string,
        # ) => {
        def set_condition_properties(search_start: Optional[str] = None,
                                     search_end: Optional[str] = None,
                                     max_duration: Optional[str] = None,
                                     columns: Optional[str] = None,
                                     sort_by: Optional[str] = None,
                                     sort_asc: Optional[str] = None):
            # dateRange.condition.range = {
            #   start: toNumber(searchStart),
            #   end: toNumber(searchEnd),
            # };
            date_range['condition']['range'] = {
                'start': int(search_start) if search_start is not None else None,
                'end': int(search_end) if search_end is not None else None,
            }
            #
            # if (columns) {
            if columns is not None:
                # dateRange.condition.columns = columns.split(',');
                date_range['condition']['columns'] = columns.split(',')
                # dateRange.condition.sortBy = sortBy;
                date_range['condition']['sortBy'] = sort_by
                # dateRange.condition.sortAsc = sortAsc === 'true';
                date_range['condition']['sortAsc'] = sort_asc.lower() == 'true'
                # dateRange.condition.isCapsuleFromTable = true;
                date_range['condition']['isCapsuleFromTable'] = True
            # } else {
            else:
                # dateRange.condition.isCapsuleFromTable = false;
                date_range['condition']['isCapsuleFromTable'] = False
            # }
            #
            # if (maxDuration) {
            if max_duration is not None:
                # const maximumDuration = splitDuration(maxDuration);
                maximum_duration = _split_duration(max_duration)
                # if (!maximumDuration) {
                if maximum_duration is None:
                    # throw new Error(`Could not parse ${maxDuration} as a maximum duration`);
                    raise ValueError(f'Could not parse {max_duration} as a maximum duration')
                # }
                #
                # if (!_.includes(DURATION_SCALAR_UNITS, maximumDuration.units)) {
                if maximum_duration['units'] not in DURATION_SCALAR_UNITS:
                    # throw new Error(
                    # `Invalid maximum duration unit ${maximumDuration.units} in ${maxDuration} as a maximum duration`,
                    # );
                    raise ValueError(
                        f'Invalid maximum duration unit {maximum_duration["units"]} in {max_duration} as a maximum '
                        f'duration')
                # }
                #
                # dateRange.condition.maximumDuration = maximumDuration;
                date_range['condition']['maximumDuration'] = maximum_duration

            # }

        # };
        #
        # // auto enabled
        # const setOffsetProperties = (offset = '1') => {
        def set_offset_properties(offset='1'):
            # const pick = toNumber(offset) ?? 1;
            pick = int(offset) if offset is not None else 1
            # if (pick === 1) {
            if pick == 1:
                # dateRange.condition.strategy = CAPSULE_SELECTION.STRATEGY.CLOSEST_TO;
                date_range['condition']['strategy'] = CAPSULE_SELECTION.STRATEGY.CLOSEST_TO
                # dateRange.condition.reference = CAPSULE_SELECTION.REFERENCE.START;
                date_range['condition']['reference'] = CAPSULE_SELECTION.REFERENCE.START
                # dateRange.condition.offset = 1;
                date_range['condition']['offset'] = 1
            # } else if (pick > 1) {
            elif pick > 1:
                # dateRange.condition.strategy = CAPSULE_SELECTION.STRATEGY.OFFSET_BY;
                date_range['condition']['strategy'] = CAPSULE_SELECTION.STRATEGY.OFFSET_BY
                # dateRange.condition.reference = CAPSULE_SELECTION.REFERENCE.START;
                date_range['condition']['reference'] = CAPSULE_SELECTION.REFERENCE.START
                # dateRange.condition.offset = pick - 1;
                date_range['condition']['offset'] = pick - 1
            # } else if (pick === -1) {
            elif pick == -1:
                # dateRange.condition.strategy = CAPSULE_SELECTION.STRATEGY.CLOSEST_TO;
                date_range['condition']['strategy'] = CAPSULE_SELECTION.STRATEGY.CLOSEST_TO
                # dateRange.condition.reference = CAPSULE_SELECTION.REFERENCE.END;
                date_range['condition']['reference'] = CAPSULE_SELECTION.REFERENCE.END
                # dateRange.condition.offset = 1;
                date_range['condition']['offset'] = 1
            # } else if (pick < -1) {
            elif pick < -1:
                # dateRange.condition.strategy = CAPSULE_SELECTION.STRATEGY.OFFSET_BY;
                date_range['condition']['strategy'] = CAPSULE_SELECTION.STRATEGY.OFFSET_BY
                # dateRange.condition.reference = CAPSULE_SELECTION.REFERENCE.END;
                date_range['condition']['reference'] = CAPSULE_SELECTION.REFERENCE.END
                # dateRange.condition.offset = Math.abs(pick) - 1;
                date_range['condition']['offset'] = abs(pick) - 1
            # }

        # };
        #
        # if (!dateRange.auto.enabled && !_.includes(dateRangeOutput.formula, 'Offset')) {
        if not date_range['auto']['enabled'] and 'Offset' not in date_range_output.formula:
            # // Fixed, with condition capsule selected from table
            # const fixedMatchPattern =
            #   /^\s*\/\/ searchStart=(.*?)ms\n\s*\/\/ searchEnd=(.*?)ms\n\s*\/\/ columns=(.*?)\n\s*\/\/ sortBy=(
            #  .*?)\n\s*\/\/ sortAsc=(.*?)\n\s* capsule\(.+\)$/m;
            fixed_match_pattern = re.compile(
                r'^\s*// searchStart=(.*?)ms\n\s*// searchEnd=(.*?)ms\n\s*// columns=(.*?)\n\s*// '
                r'sortBy=(.*?)\n\s*// sortAsc=(.*?)\n\s* capsule\(.+\)$', re.MULTILINE)
            # const fixedMatches = dateRangeOutput.formula?.match(fixedMatchPattern);
            fixed_matches = fixed_match_pattern.search(date_range_output.formula)
            # if (dateRangeOutput.condition?.id && fixedMatches) {
            if date_range_output.condition is not None and fixed_matches is not None:
                # const [notUsed, parsedStart, parsedEnd, columns, sortBy, sortAsc] = fixedMatches;
                parsed_start, parsed_end, columns, sort_by, sort_asc = fixed_matches.groups()
                # setConditionProperties(parsedStart, parsedEnd, undefined, columns, sortBy, sortAsc);
                set_condition_properties(parsed_start, parsed_end, None, columns, sort_by, sort_asc)
                # formulaFormValid = true;
                formula_form_valid = True
            # }
        # } else if (!dateRange.auto.enabled && dateRangeOutput.formula?.includes('Offset')) {
        elif (not date_range['auto']['enabled'] and date_range_output.formula is not None and 'Offset' in
              date_range_output.formula):
            # // Fixed, with selected relative capsule
            # const fixedConfigMatchPattern =
            #   /^\s*\/\/ searchStart=(.*?)ms\n\s*\/\/ searchEnd=(.*?)ms\n\s*\/\/ capsuleOffset=(.*?)\n\s*\/\/
            #  maxDuration=(.*?)\n\s*capsule\(.+\)$/m;
            fixed_config_match_pattern = re.compile(
                r'^\s*// searchStart=(.*?)ms\n\s*// searchEnd=(.*?)ms\n\s*// capsuleOffset=(.*?)\n\s*// '
                r'maxDuration=(.*?)\n\s*capsule\(.+\)$', re.MULTILINE)
            # const fixedMatches = dateRangeOutput.formula.match(fixedConfigMatchPattern);
            fixed_matches = fixed_config_match_pattern.search(date_range_output.formula)
            # if (dateRangeOutput.condition?.id && fixedMatches) {
            if (date_range_output.condition is not None and date_range_output.condition.get('id') is not None and
                    fixed_matches is not None):
                # const [notUsed, parsedStart, parsedEnd, capsuleOffset, parsedMaxDuration] = fixedMatches;
                parsed_start, parsed_end, capsule_offset, parsed_max_duration = fixed_matches.groups()
                # setConditionProperties(parsedStart, parsedEnd, parsedMaxDuration);
                set_condition_properties(parsed_start, parsed_end, parsed_max_duration)
                # setOffsetProperties(capsuleOffset);
                set_offset_properties(capsule_offset)
                # formulaFormValid = true;
                formula_form_valid = True
            # }
        # }
        #
        # // Auto-update with condition
        # const autoConditionMatchPattern =
        #   /^\$condition(.removeLongerThan\(.+?\))?.setCertain\(\).toGroup\(capsule\(.+\)\).pick\((-?[1-9]+[0-9]*)\)$/;
        auto_condition_match_pattern = re.compile(
            r'^\$condition(.removeLongerThan\(.+?\))?.setCertain\(\).toGroup\(capsule\(.+\)\).pick\((-?[1-9]+[0-9]*)\)$'
        )
        # const autoMatches = dateRangeOutput.formula?.match(autoConditionMatchPattern);
        auto_matches = (auto_condition_match_pattern.search(date_range_output.formula)
                        if date_range_output.formula is not None else None)
        # if (dateRangeOutput.condition?.id && autoMatches) {
        if date_range_output.condition is not None and auto_matches is not None:
            # // Closest to Start =>  $condition.setCertain().toGroup(capsule(start, end)).pick(1)
            # // Closest to End => $condition.setCertain().toGroup(capsule(start, end)).pick(-1)
            # // Offset by 1 from Start => $condition.setCertain().toGroup(capsule(start, end)).pick(2)
            # // Offset by 2 from Start => $condition.setCertain().toGroup(capsule(start, end)).pick(3)
            # // Offset by 1 from End => $condition.setCertain().toGroup(capsule(start, end)).pick(-2)
            parsed_pick = None
            parsed_maximum_duration = None
            if date_range_output.formula is not None:
                # const parsedPick = dateRangeOutput.formula?.match(/.*\.pick\((-?[1-9]+[0-9]*)\)$/)[1];
                parsed_pick = re.search(r'.*\.pick\((-?[1-9]+[0-9]*)\)$', date_range_output.formula).group(1)
                # const parsedMaximumDuration = dateRangeOutput.formula?.match(/removeLongerThan\((.+)\)/)?.[1];
                maximum_duration_match = re.search(r'removeLongerThan\((.+)\)', date_range_output.formula)
                if maximum_duration_match is not None:
                    parsed_maximum_duration = maximum_duration_match.group(1)
            # setConditionProperties(undefined, undefined, parsedMaximumDuration);
            set_condition_properties(None, None, parsed_maximum_duration)
            # setOffsetProperties(parsedPick);
            set_offset_properties(parsed_pick)
            # formulaFormValid = true;
            formula_form_valid = True
        # }
        #
        # // Auto-update, condition and non-condition
        # if (dateRange.auto.enabled) {
        if date_range['auto']['enabled']:
            # const background = dateRangeOutput.isBackground;
            background = date_range_output.is_background
            # const cronSchedule = dateRangeOutput.cronSchedule;
            cron_schedule = date_range_output.cron_schedule
            # const duration = extractDurationFromFormula(dateRangeOutput.formula);
            duration = _extract_duration_from_formula(date_range_output.formula)
            # const offsetAndDirection = extractOffsetAndDirectionFromFormula(dateRangeOutput.formula ?? '');
            offset_and_direction = _extract_offset_and_direction_from_formula(date_range_output.formula or '')
            # const offset = offsetAndDirection[0];
            offset = offset_and_direction[0]
            # const offsetDirection = offsetAndDirection[1];
            offset_direction = offset_and_direction[1]
            #
            # dateRange.auto = {
            date_range['auto'] = {
                # enabled: true,
                'enabled': True,
                # duration,
                'duration': duration,
                # offset,
                'offset': offset,
                # offsetDirection,
                'offsetDirection': offset_direction,
                # background,
                'background': background,
                # cronSchedule,
                'cronSchedule': cron_schedule,
                # };
            }
        # }
    # } catch (e) {
    except Exception as e:
        # logWarn(e);
        # formulaFormValid = false;
        formula_form_valid = False
    # }
    #
    # if (!formulaFormValid) {
    if not formula_form_valid:
        # logWarn(`Failed to parse date range formula "${dateRangeOutput.formula}" [${dateRangeOutput.id}]`);
        # dateRange.irregularFormula = dateRangeOutput.formula;
        date_range['irregularFormula'] = date_range_output.formula
    # }
    # return dateRange;
    return date_range
    # };


def compute_capsule_offset(condition):
    strategy = condition["strategy"]
    reference = condition["reference"]
    offset = condition.get("offset", 1)
    offset_value = 1 if strategy == "closestTo" else int(offset) + 1
    sign_value = 1 if reference == "start" else -1
    return offset_value * sign_value


def create_date_range_formula(date_range: dict) -> str:
    # if (!dateRange.auto.enabled && !dateRange.condition.id) {
    # return `capsule(${dateRange.range.start}ms, ${dateRange.range.end}ms)`;
    # }
    if (not date_range['auto']['enabled'] and
            (date_range['condition'] is None or date_range['condition'].get('id') is None)):
        # Fixed range, no condition
        return f"capsule({date_range['range']['start']}ms, {date_range['range']['end']}ms)"

    # const capsuleOffset = computeCapsuleOffset(dateRange.condition);
    capsule_offset = compute_capsule_offset(date_range['condition'])

    # const maximumDuration = dateRange.condition?.maximumDuration
    # ? `${dateRange.condition.maximumDuration.value}${dateRange.condition.maximumDuration.units}`
    # : '';
    maximum_duration = (
        f"{date_range['condition']['maximumDuration']['value']}{date_range['condition']['maximumDuration']['units']}"
        if (date_range['condition'] is not None and date_range['condition'].get('maximumDuration') is not None) else '')

    if (not date_range['auto']['enabled'] and date_range['condition'].get('id') is not None
            and date_range['condition'].get('isCapsuleFromTable', False)):
        # Fixed range, condition
        # For this configuration, we save the search range and other parameters used to find the capsule in a formula
        # comment, so that we can present it back to the user in the UI when editing the dateRange.
        return (f"""// searchStart={date_range['condition']['range']['start']}ms
                // searchEnd={date_range['condition']['range']['end']}ms
                // columns={','.join(date_range['condition']['columns'])}
                // sortBy={date_range['condition']['sortBy']}
                // sortAsc={str(date_range['condition']['sortAsc']).lower()}
                capsule({date_range['range']['start']}ms, {date_range['range']['end']}ms)""")
    elif (not date_range['auto']['enabled'] and date_range['condition'].get('id') is not None
          and not date_range['condition'].get('isCapsuleFromTable', False)):
        return (f"""// searchStart={date_range['condition']['range']['start']}ms
                // searchEnd={date_range['condition']['range']['end']}ms
                // capsuleOffset={capsule_offset}
                // maxDuration={maximum_duration}
                capsule({date_range['range']['start']}ms, {date_range['range']['end']}ms)""")
    else:
        # Auto-update range
        sign = '+' if date_range['auto']['offsetDirection'] == OFFSET_DIRECTION.FUTURE else '-'
        offset = f"{date_range['auto']['offset']['value']}{date_range['auto']['offset']['units']}"
        duration = f"{date_range['auto']['duration']}ms"
        range_formula = f"capsule($now {sign} {offset} - {duration}, $now {sign} {offset})"

    if date_range['condition'].get('id') is not None:
        maximum_duration_snippet = f".removeLongerThan({maximum_duration})" if maximum_duration else ''
        return f"$condition{maximum_duration_snippet}.setCertain().toGroup({range_formula}).pick({capsule_offset})"
    else:
        return range_formula


def _split_duration(duration_string):
    value = None
    units = None

    if isinstance(duration_string, str):
        value = re.search(r'[-+]?[0-9]*\.?[0-9]+', duration_string).group(0)

        if value:
            units = re.search(r'[a-zA-Z]+', duration_string).group(0)

            if units:
                return {'value': value, 'units': units}


def _extract_duration_from_formula(formula: str) -> Optional[int]:
    match = re.search(r'\$now\s*[+-]\s*(.*?)\s*[+-]\s*([\d.]+[a-z]+),', formula, re.IGNORECASE)
    duration_match = match.group(2) if match is not None else None
    return int(_parse_duration(duration_match).total_seconds() * 1000) if duration_match is not None else None


# export function parseDuration(durationInput: string): Duration {
def _parse_duration(duration_input: str) -> datetime.timedelta:
    # const invalidDuration = moment.duration(0);
    invalid_duration = datetime.timedelta(0)
    # const trimmedInput = _.trim(durationInput);
    trimmed_input = duration_input.strip()
    # if (isNaN(Number(trimmedInput[0]))) {
    if not trimmed_input[0].isnumeric():
        # return invalidDuration;
        return invalid_duration
    # }
    #
    # const patterns = i18next.t('DURATION_PATTERNS', { returnObjects: true });
    patterns = {
        'DURATION_PATTERNS': {
            'years': ['y', 'yr', 'yrs', 'year', 'years'],
            'months': ['mo', 'mos', 'month', 'months'],
            'weeks': ['w', 'wk', 'wks', 'week', 'weeks'],
            'days': ['d', 'day', 'days'],
            'hours': ['h', 'hr', 'hrs', 'hour', 'hours'],
            'minutes': ['m', 'min', 'mins', 'minute', 'minutes'],
            'seconds': ['s', 'sec', 'secs', 'second', 'seconds'],
            'milliseconds': ['ms', 'millisecond', 'milliseconds'],
        }
    }
    # const DURATION_CONVERSIONS = {
    DURATION_CONVERSIONS = {
        'years': 31536000000,
        'months': 2592000000,
        'weeks': 604800000,
        'days': 86400000,
        'hours': 3600000,
        'minutes': 60000,
        'seconds': 1000,
        'milliseconds': 1,
    }
    #
    # const matches: string[] = [];
    matches = []

    duration_in_ms = 0
    for key in DURATION_CONVERSIONS.keys():
        # for (const pattern of patterns[key]) {
        for pattern in patterns['DURATION_PATTERNS'][key]:
            # const regex = new RegExp(`((\\d+\\.\\d+)|\\d+)\\s?(${pattern}(?=\\s|\\d|$))`, 'gi');
            regex = re.compile(f"((\\d+\\.\\d+)|\\d+)\\s?({pattern}(?=\\s|\\d|$))", re.IGNORECASE)
            # for (const match of durationInput.matchAll(regex)) {
            for match in [[m.group(0)] + list(m.groups()) for m in regex.finditer(trimmed_input)]:
                # if (!_.includes(matches, match[0])) {
                if match[0] not in matches:
                    # matches.push(match[0]);
                    matches.append(match[0])
                    # return total + parseFloat(match[1]) * multiplier;
                    duration_in_ms += float(match[1]) * DURATION_CONVERSIONS[key]
                # }
            # }
        # }
        #
        # return total;
    # },

    # const durationInSeconds: number = _.reduce(
    #     (total, multiplier, key) => {
    #     DURATION_CONVERSIONS,
    #     0,
    # );

    #
    # if (_.isEmpty(matches)) {
    if not matches:
        # return invalidDuration;
        return invalid_duration
    # } else {
    else:
        # return moment.duration(durationInSeconds, 'seconds');
        return datetime.timedelta(milliseconds=duration_in_ms)
    # }


# }


#   export const extractOffsetAndDirectionFromFormula = (formula = '') => {
def _extract_offset_and_direction_from_formula(formula: str = '') -> Tuple[Dict[str, int], str]:
    # const formulaOffset = formula.match(/[+-](.*?)[-](.*?)(\$now)(.*)/)[4];
    formula_offset = re.search(r'[+-](.*?)[-](.*?)(\$now)(.*)', formula).group(4)
    # const value = toNumber(formulaOffset.match(/(\d+)(\w+)/)[1]);
    value = int(re.search(r'(\d+)(\w+)', formula_offset).group(1))
    # const units = formulaOffset.match(/(\d+)(\w+)/)[2];
    units = re.search(r'(\d+)(\w+)', formula_offset).group(2)
    # const offset = { value, units };
    offset = {'value': value, 'units': units}
    #
    # const offsetDirection = formulaOffset.match(/[+-]/)[0] === '-' ? OFFSET_DIRECTION.PAST : OFFSET_DIRECTION.FUTURE;
    offset_direction = (OFFSET_DIRECTION.PAST if re.search(r'[+-]', formula_offset).group(0) == '-' else
                        OFFSET_DIRECTION.FUTURE)
    #
    # return [offset, offsetDirection];
    return offset, offset_direction
# };



from __future__ import annotations

import setuptools
from setuptools.command.sdist import sdist as _sdist

class CustomSdistCommand(_sdist):
    def make_distribution(self):
        # CRAB-45784: PEP 625 no longer allows distributions (the .tar.gz file) to include non-normalized chars
        # ('-' in this case) in the file name. Replace the name with an underscore only for the distribution without
        # altering SPy's package name.
        # TODO CRAB-35238: This can be removed when setuptools is upgraded to v69.3.0 or later.
        self.distribution.metadata.name = self.distribution.metadata.name.replace("-", "_")
        super().make_distribution()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="seeq-spy",
    version="196.4",
    author="Seeq Corporation",
    author_email="support@seeq.com",
    description="Easy-to-use Python interface for Seeq",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.seeq.com",
    project_urls={
        'Documentation': 'https://python-docs.seeq.com/',
        'Changelog': 'https://python-docs.seeq.com/changelog.html'
    },
    packages=setuptools.find_namespace_packages(exclude=['seeq.sdk', 'seeq.sdk.*']),
    include_package_data=True,
    install_requires=[
        'Deprecated >= 1.2.6',
        'numpy >= 1.16.4',
        'pandas >= 1.0.0',
        'python-dateutil >= 2.7.0',
        'pytz >= 2020.1',
        'requests >= 2.22.0',
        'urllib3 >= 1.21.1',
        'dataclasses >= 0.7; python_version == "3.6"'  # This is a polyfill for dataclasses in Python 3.6.
    ],
    extras_require={
        # This should remain updated with the dependency_mapping in _dependencies.py
        'widgets': [
            'ipython >= 7.6.1',
            'ipywidgets >= 7.5.1',
            'matplotlib >= 3.1.1',
            'seeq-data-lab-env-mgr >= 0.1.0',
        ],
        'templates': [
            'beautifulsoup4 >= 4.8.0',
            'chevron >= 0.13.1',
            'Mako >= 1.1.0',
            'Markdown >= 3.3.4',
            'matplotlib >= 3.1.1',
        ],
        'jobs': [
            'cron-descriptor >= 1.2.24',
            'nbconvert >= 6.4.4',
            'nbformat >= 5.2.0',
            'recurrent >= 0.4.0',
        ],
        'jupyter': [
            'ipylab >= 0.5.2',
            'ipython >= 7.6.1',
            'ipywidgets >= 7.5.1',
            'jupyterlab >= 3.0.0',
            'nbconvert >= 6.4.4',
            'nbformat >= 5.2.0',
            'notebook >= 6.0.0',
            'psutil >= 5.9.0',
            'setuptools >= 65.0.0',
        ],
        'all': [
            'beautifulsoup4 >= 4.8.0',
            'chevron >= 0.13.1',
            'cron-descriptor >= 1.2.24',
            'ipylab >= 0.5.2',
            'ipywidgets >= 7.5.1',
            'jupyterlab >= 3.0.0',
            'Markdown >= 3.3.4',
            'matplotlib >= 3.1.1',
            'nbconvert >= 6.4.4',
            'nbformat >= 5.2.0',
            'notebook >= 6.0.0',
            'psutil >= 5.9.0',
            'recurrent >= 0.4.0',
            'setuptools >= 65.0.0',
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
    cmdclass={"sdist": CustomSdistCommand},
)



from __future__ import annotations

import os
import tempfile
import zipfile
from typing import Optional

from seeq.base import util
from seeq.spy import _common
from seeq.spy._errors import *
from seeq.spy._status import Status
from seeq.spy.workbooks._workbook import Workbook


def save(workbooks, folder_or_zipfile: Optional[str] = None, *, datasource_map_folder: Optional[str] = None,
         include_rendered_content: bool = False, pretty_print_html=False, overwrite: bool = False,
         errors: Optional[str] = None, quiet: Optional[bool] = None, status: Optional[Status] = None):
    """

    Saves a list of workbooks to a folder on disk from Workbook objects in
    memory.

    Parameters
    ----------
    workbooks : {Workbook, list[Workbook]}
        A Workbook object or list of Workbook objects to save.

    folder_or_zipfile : str, default os.getcwd()
        A folder or zip file on disk to which to save the workbooks. It will
        be saved as a "flat" set of subfolders, no other hierarchy will be
        created. The string must end in ".zip" to cause a zip file to be
        created instead of a folder.

    datasource_map_folder : str, default None
        Specifies a curated set of datasource maps that should accompany the
        workbooks (as opposed to the default maps that were created during the
        spy.workbooks.pull call).

    include_rendered_content : bool, default False
        If True, creates a folder called RenderedTopic within each Topic
        Document folder that includes the embedded content such that you can
        load it in an offline browser. You are required to have pulled the
        workbooks with include_rendered_content=True.

    pretty_print_html : bool, default False
        If True, this function will re-format the HTML of Topic Documents
        and Journals so that it is easy to read in a text editor. If False, the
        HTML will be written exactly as it is stored in Seeq Server. Note that
        setting this to True can cause some minor deviations in rendering after
        a round-trip (i.e. pull/save/load/push).

    overwrite : bool, default False
        If True, will overwrite any existing folder or zip file.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame (errors='catalog' must be combined with
        status=<Status object>).

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedence.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed.
    """
    _common.validate_argument_types([
        (workbooks, 'workbooks', (Workbook, list)),
        (folder_or_zipfile, 'folder_or_zipfile', str),
        (datasource_map_folder, 'datasource_map_folder', str),
        (include_rendered_content, 'included_rendered_content', bool),
        (pretty_print_html, 'pretty_print_html', bool),
        (overwrite, 'overwrite', bool),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status)
    ])

    status = Status.validate(status, None, quiet, errors)

    if not folder_or_zipfile:
        folder_or_zipfile = os.getcwd()

    folder_or_zipfile = os.path.normpath(folder_or_zipfile)

    try:
        if not isinstance(workbooks, list):
            workbooks = [workbooks]

        if folder_or_zipfile is None:
            folder_or_zipfile = os.getcwd()

        if not os.path.isabs(folder_or_zipfile):
            folder_or_zipfile = util.safe_abspath(folder_or_zipfile)

        zip_it = folder_or_zipfile.lower().endswith('.zip')

        datasource_maps = None if datasource_map_folder is None else Workbook.load_datasource_maps(
            datasource_map_folder)

        if util.safe_isfile(folder_or_zipfile):
            if overwrite:
                util.safe_remove(folder_or_zipfile)
            else:
                raise SPyRuntimeError(
                    f'"{folder_or_zipfile}" already exists. Use overwrite=True to overwrite.')

        save_folder = None
        try:
            save_folder = tempfile.mkdtemp() if zip_it else folder_or_zipfile

            for workbook in workbooks:  # type: Workbook
                if not isinstance(workbook, Workbook):
                    raise SPyTypeError('workbooks argument must be a list of Workbook objects')

                workbook_folder_name = '%s (%s)' % (workbook.name, workbook.id)
                workbook_folder = os.path.join(save_folder, util.cleanse_filename(workbook_folder_name))

                if datasource_maps is not None:
                    workbook.datasource_maps = datasource_maps

                status.update('Saving to "%s"' % workbook_folder, Status.RUNNING)
                workbook.save(workbook_folder, include_rendered_content=include_rendered_content,
                              pretty_print_html=pretty_print_html, overwrite=overwrite)

            if zip_it:
                status.update('Zipping "%s"' % folder_or_zipfile, Status.RUNNING)
                util.safe_makedirs(os.path.dirname(folder_or_zipfile), exist_ok=True)
                with zipfile.ZipFile(util.handle_long_filenames(folder_or_zipfile),
                                     "w", zipfile.ZIP_DEFLATED) as z:
                    for root, dirs, files in util.safe_walk(save_folder):
                        for file in files:
                            filename = os.path.join(root, file)
                            if util.safe_isfile(filename):  # regular files only
                                archive_name = os.path.join(util.safe_relpath(root, save_folder), file)
                                _common.print_output('Archiving %s' % archive_name)
                                # noinspection PyTypeChecker
                                z.write(filename, archive_name)

        finally:
            if save_folder and zip_it:
                util.safe_rmtree(save_folder)

        status.update('Success', Status.SUCCESS)

    except KeyboardInterrupt:
        status.update('Save canceled', Status.CANCELED)



from __future__ import annotations

from seeq.spy.workbooks.job.data._pull import pull, manifest, expand, add, remove, calculation
from seeq.spy.workbooks.job.data._push import push
from seeq.spy.workbooks.job.data._redo import redo

__all__ = ['pull', 'push', 'manifest', 'expand', 'add', 'remove', 'redo', 'calculation']



from __future__ import annotations

import textwrap
from abc import ABC, abstractmethod
from collections.abc import MutableSequence
from typing import Union, List, Dict, Any, cast, TypeVar, Generic

import pandas as pd
from seeq.spy import _common
from seeq.spy._common import docstring_parameter
from seeq.spy._errors import *
from seeq.spy.workbooks._context_switchable import ContextSwitchable

ColsT = TypeVar('ColsT', bound='Columns')


# refer client/packages/webserver/app/src/tableBuilder
class TableToolbar(ContextSwitchable, Generic[ColsT]):
    class CONSTANTS:
        TABLE_STORE_NAME = 'sqTableBuilderStore'
        SIMPLE_TABLE_MODE = 'simple'
        CONDITION_TABLE_MODE = 'condition'

        TRANSPOSE = 'isTransposed'
        STRIPED = 'isTableStriped'
        AUTO_GROUP_COLUMN = 'autoGroupColumn'

        DEFAULT_TABLE_STATE = {
            'headers': {
                'condition': {
                    'type': 'startEnd',
                    'format': 'lll'
                },
                'simple': {
                    'type': 'startEnd',
                    'format': 'lll'
                }
            },
            'columns': {
                'condition': [],
                'simple': []
            }
        }

    TABLE_MODE = None

    def __init__(self, parent):
        self._parent = parent
        super().__init__(parent)
        self._columns = None

    @property
    def view(self) -> str:
        """
        The current view for the workstep. Valid values are

        ============ =========================================
        View         Result
        ============ =========================================
        Trend        Show the time-series trend view (default)
        Scatter Plot Show the scatter plot view
        Treemap      Show the treemap view
        Table        Show the table view
        ============ =========================================
        """
        return self._getter_workstep.view

    @view.setter
    def view(self, value: str):
        self._setter_workstep.view = value

    @property
    @docstring_parameter(CONSTANTS.SIMPLE_TABLE_MODE, CONSTANTS.CONDITION_TABLE_MODE)
    def table_mode(self) -> str:
        """
        The current table mode for the workstep.
        Can either be `{0}` or `{1}`

        """
        return self._get_table_mode()

    @table_mode.setter
    def table_mode(self, value: str):
        self._set_table_mode(value)

    @property
    def transpose(self) -> bool:
        """
        Boolean value indicating if the table is transposed.
        In Condition Table mode, a value of `True` produces a table where rows correspond to capsules
        In Simple Table mode, a value of `True` produces a table where rows correspond to item statistics

        Can either be `True` or `False`
        """
        return self._get_transpose()

    @transpose.setter
    def transpose(self, value: bool):
        self._set_transpose(value)

    @property
    def striped(self) -> bool:
        """
        Boolean value indicating if every alternate row in the table is shaded.
        Can either be `True` or `False`
        """
        return self._get_striped()

    @striped.setter
    def striped(self, value: bool):
        self._set_striped(value)

    @property
    def group(self):
        """
        Boolean value indicating if grouping is enabled in the table.
        Can either be `True` or `False`
        """
        return self._get_group()

    @group.setter
    def group(self, value: bool):
        self._set_group(value)

    @property
    def columns(self) -> ColsT:
        """
        Column configuration for Condition Tables. Use methods on this object to add or remove columns from the table.
        """
        return self._columns

    @columns.setter
    def columns(self, value: list[TableColumn]):
        self.columns._set_columns(value)

    @staticmethod
    def _get_table_store(workstep):
        workstep_stores = workstep.get_workstep_stores()
        return _common.get(
            workstep_stores, TableToolbar.CONSTANTS.TABLE_STORE_NAME,
            default=TableToolbar.CONSTANTS.DEFAULT_TABLE_STATE,
            assign_default=True
        )

    @staticmethod
    def get_config_for_table_mode(workstep, attribute: str, table_mode: str, default=None):
        table_store = TableToolbar._get_table_store(workstep)
        table_config = _common.get(table_store, attribute, default=dict(), assign_default=False)
        return table_config.get(table_mode, default)

    @staticmethod
    def set_config_for_table_mode(workstep, attribute: str, table_mode: str, value):
        table_store = TableToolbar._get_table_store(workstep)
        table_config = _common.get(table_store, attribute, default=dict(), assign_default=True)
        table_config[table_mode] = value

    def _get_table_mode(self) -> str:
        return self._getter_workstep.table_mode

    def _set_table_mode(self, value: str):
        self._setter_workstep.table_mode = value

    def _get_transpose(self) -> bool:
        return TableToolbar.get_config_for_table_mode(self._getter_workstep, self.CONSTANTS.TRANSPOSE,
                                                      self.TABLE_MODE, False)

    def _set_transpose(self, value: bool):

        if isinstance(value, bool):
            TableToolbar.set_config_for_table_mode(self._setter_workstep, self.CONSTANTS.TRANSPOSE,
                                                   self.TABLE_MODE, value)
        else:
            raise SPyTypeError("'transpose' must be a boolean")

    def _get_striped(self) -> bool:
        return TableToolbar.get_config_for_table_mode(self._getter_workstep, self.CONSTANTS.STRIPED, self.TABLE_MODE,
                                                      False)

    def _set_striped(self, value: bool):
        if isinstance(value, bool):
            TableToolbar.set_config_for_table_mode(self._setter_workstep, self.CONSTANTS.STRIPED,
                                                   self.TABLE_MODE, value)
        else:
            raise SPyTypeError("'striped' must be a boolean")

    def _get_group(self) -> bool:
        # grouping is determined by whether there is any value for the table mode for autoGroupColumn
        auto_group_column = TableToolbar.get_config_for_table_mode(self._getter_workstep,
                                                                   self.CONSTANTS.AUTO_GROUP_COLUMN,
                                                                   self.TABLE_MODE,
                                                                   None)

        return auto_group_column is not None

    def _set_group(self, value: bool):
        # grouping is enabled by setting a value for the table mode for autoGroupColumn
        if isinstance(value, bool):
            if value:
                TableToolbar.set_config_for_table_mode(self._setter_workstep, self.CONSTANTS.AUTO_GROUP_COLUMN,
                                                       self.TABLE_MODE, {})
            else:
                TableToolbar.set_config_for_table_mode(self._setter_workstep, self.CONSTANTS.AUTO_GROUP_COLUMN,
                                                       self.TABLE_MODE, None)
        else:
            raise SPyTypeError("'group' must be a boolean")


class SimpleTableToolbar(TableToolbar['SimpleTableColumns']):
    """
    Exposes configuration options for Simple Tables in a workbench Analysis.

    Methods are named based on a table with Transpose set to `False`: rows are items, columns are statistics or
    properties.
    """
    TABLE_MODE = TableToolbar.CONSTANTS.SIMPLE_TABLE_MODE

    def __init__(self, parent):
        super().__init__(parent)
        self._columns = SimpleTableColumns(parent)


class ConditionTableToolbar(TableToolbar['ConditionTableColumns']):
    """
    Exposes configuration options for Condition Tables in a workbench Analysis.

    Methods are named based on a table with Transpose set to `True`: rows are capsules, columns are statistics or
    properties.
    """
    TABLE_MODE = TableToolbar.CONSTANTS.CONDITION_TABLE_MODE

    def __init__(self, parent):
        super().__init__(parent)
        self._columns = ConditionTableColumns(parent)

    def add_row_column_labels(self) -> ConditionTableRow:
        """
        Enable and return the row that displays column labels in the table.
        Once returned, the row can be manipulated to adjust the appearance of the column labels.
        """
        return self.columns.add_row_column_labels()

    def remove_row_column_labels(self):
        """
        Hide the row that displays column labels in the table.
        """
        self.columns.remove_row_column_labels()

    def add_row_unit_of_measure(self) -> ConditionTableRow:
        """
        Enable and return the row that displays units of measure in the table.
        When the row is present, units will be moved from within each cell to the row.
        Once returned, this row can be manipulated to adjust the appearance of the units of measure.
        """
        return self.columns.add_row_unit_of_measure()

    def remove_row_unit_of_measure(self):
        """
        Hide the row that displays units of measure in the table.
        """
        self.columns.remove_row_unit_of_measure()

    def add_row_custom_text(self, column_values: Union[List[str], Dict[str, str]]) -> ConditionTableRowCustomText:
        """
        Add and return a row that displays custom text in the table.
        Once returned, the row can be manipulated to adjust the appearance of the custom text.

        Parameters
        ----------
        column_values: Union[List[str], Dict[str, str]]
            List of string values to display in the row or a dictionary of column keys and values
        """
        return self.columns.add_row_custom_text(column_values)

    def remove_row_custom_text(self):
        """
        Remove all rows that displays custom text in the table.
        """
        self.columns.remove_row_custom_text()

    def _column_labels_enabled(self) -> bool:
        return ConditionTableRow.CONSTANTS.COLUMN_LABELS in self.columns._get_columns()

    def _unit_of_measure_enabled(self) -> bool:
        return ConditionTableRow.CONSTANTS.UNIT_OF_MEASURE in self.columns._get_columns()

    def __repr__(self):
        return (
            f"Condition Table Configuration:\n"
            f"  - Table Mode: {self.table_mode}\n"
            f"  - Transpose: {self.transpose}\n"
            f"  - Striped: {self.striped}\n"
            f"  - Group: {self.group}\n"
            f"  - Column Labels: {self._column_labels_enabled()}\n"
            f"  - Units of Measure Row: {self._unit_of_measure_enabled()}\n"
            f"{textwrap.indent(f'- {repr(self.columns)}', prefix='  ')}"
        )


class TableColumn(ABC, ContextSwitchable):
    class CONSTANTS:
        # refer client/packages/webserver/app/src/core/TextFormatterPopover.molecule.tsx
        TEXT_ALIGN_OPTIONS = {'left', 'center', 'right'}
        TEXT_STYLE_OPTIONS = {'bold', 'italic', 'underline', 'line-through', 'overline'}

        # refer AgGridAggregationFunctions in  client/packages/webserver/app/src/tableBuilder/tableBuilder.constants.ts
        GROUPING_AGGREGATION_FUNCTIONS = {
            'none',
            'sum',
            'min',
            'max',
            'count',
            'avg',
            'range',
            'stdDev',
            'first',
            'last',
        }

        SIGNAL_STATISTICS = _common.get_trend_signal_statistic_function_map()
        SIGNAL_STATISTICS_TO_NAME = {v: k for k, v in SIGNAL_STATISTICS.items()}

    @property
    @abstractmethod
    def is_row(self) -> bool:
        pass

    @abstractmethod
    def _get_column_definition(self) -> dict:
        pass


class ConditionTableColumn(TableColumn):
    """
    The SPy representation of a column in a condition table.
    """

    class CONSTANTS(TableColumn.CONSTANTS):
        pass

    def __init__(self, column, parent):
        super().__init__(parent)
        self._key = column.get('key')

    def __repr__(self):
        return (f"{self.key}:\n" +
                textwrap.indent("".join([
                    f"- Key: {self.key}\n",
                    f"- Type: Property\n",
                    f"- Header: {self.header}\n" if self.header else "",
                    f"- Grouped: {self.grouping}\n",
                    f"- Group Order: {self.group_order}\n" if self.grouping else "",
                    f"- Aggregation Function: {self.aggregation_function}\n" if self.aggregation_function else ""
                ]), prefix="    "))

    def __eq__(self, other: Union[ConditionTableColumn, str]):
        # allow comparisons against strings for convenience
        if isinstance(other, str):
            return self.key == other

        return self.key == other.key

    def validate(self):
        if self.key is None:
            raise SPyValueError('Key must be provided')

    def _get_column_definition(self) -> dict:
        condition_columns = ConditionTableColumns.get_column_store(self._getter_workstep)
        return next(column for column in condition_columns if column['key'] == self.key)

    def _get_column_attribute(self, key: str, default=None) -> Any:
        return _common.get(self._get_column_definition(), key, default=default)

    def _set_column_attribute(self, key: str, value: Any):
        condition_columns = ConditionTableColumns.get_column_store(self._setter_workstep)
        column = next(column for column in condition_columns if column['key'] == self.key)
        # column attributes never read False or empty lists, the key is just removed
        if value or value == 0:
            column[key] = value
        else:
            column.pop(key, None)

    @property
    def is_row(self) -> bool:
        """
        Whether the column is displayed as a row in the table.
        """
        return False

    @property
    def key(self) -> str:
        """
        Unique identifier for the column. Must be unique within the table. Cannot be changed.
        """
        return self._key

    @key.setter
    def key(self, value: str):
        raise SPyValueError("'key' is read-only")

    @property
    def type(self) -> str:
        """
        The type of the column. Cannot be changed.
        """
        return self._get_column_attribute('type')

    @type.setter
    def type(self, value: str):
        raise SPyValueError("'type' is read-only")

    @property
    def header(self) -> str:
        """
        A string value that is displayed in the header of the column.
        Changing this will override what is displayed in the header for this column.
        """
        return self._get_column_attribute('header')

    @header.setter
    def header(self, value: str):
        if value is None or isinstance(value, str):
            self._set_column_attribute('header', value)
        else:
            raise SPyTypeError("'header' must be a string")

    @property
    @docstring_parameter(CONSTANTS.TEXT_ALIGN_OPTIONS)
    def header_text_align(self) -> str:
        """
        A string value indicating how the header text will be aligned.
        Must be one of: `{0}`
        """
        return self._get_column_attribute('headerTextAlign')

    @header_text_align.setter
    def header_text_align(self, value: str):
        if value is None or value in self.CONSTANTS.TEXT_ALIGN_OPTIONS:
            self._set_column_attribute('headerTextAlign', value)
        else:
            raise SPyValueError(f"'header_text_align' must be one of {self.CONSTANTS.TEXT_ALIGN_OPTIONS}")

    @property
    @docstring_parameter(CONSTANTS.TEXT_STYLE_OPTIONS)
    def header_text_style(self) -> List[str]:
        """
        A list of string values indicating the style of the header text.
        Must be a subset of: `{0}`
        """
        return self._get_column_attribute('headerTextStyle')

    @header_text_style.setter
    @docstring_parameter(CONSTANTS.TEXT_STYLE_OPTIONS)
    def header_text_style(self, value: List[str]):
        """
        A list of string values indicating the style of the header text.
        Must be a subset of: `{0}`
        """
        if value is None or all(style in self.CONSTANTS.TEXT_STYLE_OPTIONS for style in value):
            self._set_column_attribute('headerTextStyle', value)
        else:
            raise SPyValueError(f"'header_text_style' must be in {self.CONSTANTS.TEXT_STYLE_OPTIONS}")

    @property
    @docstring_parameter(CONSTANTS.TEXT_ALIGN_OPTIONS)
    def text_align(self) -> str:
        """
        A string value indicating how the values in the column will be aligned.
        Must be one of: `{0}`
        """
        return self._get_column_attribute('textAlign')

    @text_align.setter
    def text_align(self, value: str):
        if value is None or value in self.CONSTANTS.TEXT_ALIGN_OPTIONS:
            self._set_column_attribute('textAlign', value)
        else:
            raise SPyTypeError(f"'text_align' must be one of {self.CONSTANTS.TEXT_ALIGN_OPTIONS}")

    @property
    @docstring_parameter(CONSTANTS.TEXT_STYLE_OPTIONS)
    def text_style(self) -> List[str]:
        """
        A list of string values indicating the style of the values in the column.
        Must be a subset of: `{0}`
        """
        return self._get_column_attribute('textStyle')

    @text_style.setter
    def text_style(self, value: List[str]):
        if value is None or all(style in self.CONSTANTS.TEXT_STYLE_OPTIONS for style in value):
            self._set_column_attribute('textStyle', value)
        else:
            raise SPyTypeError(f"'text_style' must be in {self.CONSTANTS.TEXT_STYLE_OPTIONS}")

    @property
    def width(self) -> int:
        """
        Integer value indicating the width of the column.
        """
        return self._get_column_attribute('width')

    @width.setter
    def width(self, value: int):
        if value is None or isinstance(value, int):
            self._set_column_attribute('width', value)
        else:
            raise SPyTypeError("'width' must be an integer")

    @property
    def group_order(self) -> int:
        """
        Integer value indicating the order that columns are grouped in.
        This needs to be consistent across all columns that are grouped.
        To ensure this, always manipulate grouping through the :py:attr:`ConditionTableColumns.grouped_columns`
        property.
        """
        return self._get_column_attribute('rowGroupOrder')

    @group_order.setter
    def group_order(self, value: int):
        if value is None or isinstance(value, int):
            self._set_column_attribute('rowGroupOrder', value)
        else:
            raise SPyTypeError("group_order must be an integer")

    @property
    def grouping(self) -> bool:
        """
        Boolean value indicating if the column is grouped in the table.
        If `True`, group_order must also be set. To ensure this, always manipulate grouping through the
        :py:attr:`ConditionTableColumns.grouped_columns`
        property.
        """
        # if grouping not enabled, key isn't in the stores, so return False
        return self._get_column_attribute('grouping', False)

    @grouping.setter
    def grouping(self, value: bool):
        if isinstance(value, bool):
            self._set_column_attribute('grouping', value if value else None)
        else:
            raise SPyTypeError("'grouping' must be a boolean")

    @property
    @docstring_parameter(CONSTANTS.GROUPING_AGGREGATION_FUNCTIONS)
    def aggregation_function(self) -> str:
        """
        String value indicating how to aggregate values in this column when grouping is enabled.
        Must be one of: `{0}`
        """
        return self._get_column_attribute('aggregationFunction')

    @aggregation_function.setter
    def aggregation_function(self, value: str):
        if value is None or value in self.CONSTANTS.GROUPING_AGGREGATION_FUNCTIONS:
            self._set_column_attribute('aggregationFunction', value)
        else:
            raise SPyValueError(
                f"'aggregation_function' must be one of {self.CONSTANTS.GROUPING_AGGREGATION_FUNCTIONS}")

    @staticmethod
    def validate_definition(column_definition: dict) -> bool:
        return True


class ConditionTableColumnScorecard(ConditionTableColumn):
    @staticmethod
    def validate_definition(column_definition: dict) -> bool:
        return column_definition.get('metricId', None) is not None

    @property
    def metric_id(self) -> str:
        """
        The ID of the metric that the column is displaying.
        """
        return self._get_column_definition()['metricId']

    @metric_id.setter
    def metric_id(self, value: str):
        raise SPyValueError("'metric_id' is read-only")

    def __repr__(self):
        # lookup the name of the scorecard metric to show to user
        scorecard_name = self._getter_workstep.display_items.loc[
            self._getter_workstep.display_items['ID'] == self.metric_id, 'Name'].values[0]
        return (f"{scorecard_name}:\n" +
                textwrap.indent("".join([
                    f"- Key: {self.key}\n",
                    f"- Type: Scorecard Metric\n",
                    f"- Item ID: {self.metric_id}\n"
                    f"- Grouped: {self.grouping}\n",
                    f"- Group Order: {self.group_order}\n" if self.grouping else "",
                    f"- Aggregation Function: {self.aggregation_function}\n" if self.aggregation_function else ""
                ]), prefix="    "))


class ConditionTableRow(ConditionTableColumn):
    class CONSTANTS(ConditionTableColumn.CONSTANTS):
        COLUMN_LABELS = 'name'
        UNIT_OF_MEASURE = 'valueUnitOfMeasure'

        KEYS = [
            COLUMN_LABELS,
            UNIT_OF_MEASURE
        ]

    # these are the columns that actually display as rows in the table
    # includes column labels, custom text rows, and unit of measure rows
    # they do not accept custom text, and cannot be used for grouping
    @property
    def is_row(self) -> bool:
        return True

    @property
    def grouping(self) -> bool:
        return False

    @grouping.setter
    def grouping(self, value: bool):
        pass

    @staticmethod
    def validate_definition(column_definition: dict) -> bool:
        key = column_definition.get('key', None)
        return key in ConditionTableRow.CONSTANTS.KEYS


class ConditionTableRowCustomText(ConditionTableRow):
    @staticmethod
    def validate_definition(column_definition: dict) -> bool:
        return column_definition.get('type', None) == 'text'


class ConditionTableColumnReservedProperty(ConditionTableColumn):
    # refer CONDITION_EXTRA_COLUMNS in client/packages/webserver/app/src/tableBuilder/tableBuilder.constants.ts
    class CONSTANTS(ConditionTableColumn.CONSTANTS):
        START_KEY = 'startTime'
        END_KEY = 'endTime'
        NAME_EXPRESSION_KEY = 'nameExpression'
        ASSET_KEY = 'asset'

        KEY_MAPPING = {
            'start': START_KEY,
            'end': END_KEY,
            'name': NAME_EXPRESSION_KEY,
            'asset': ASSET_KEY
        }

        KEYS_OR_VALUES = set(KEY_MAPPING) | set(KEY_MAPPING.values())

    @staticmethod
    def validate_definition(column_definition: dict) -> bool:
        key = column_definition.get('key', None)
        return key in ConditionTableColumnReservedProperty.CONSTANTS.KEY_MAPPING.values()


class ConditionTableColumnSignalStatistic(ConditionTableColumn):
    @staticmethod
    def validate_definition(column_definition: dict) -> bool:
        return column_definition.get('statisticKey', None) is not None

    @property
    def signal_id(self) -> str:
        return self._get_column_definition()['signalId']

    @signal_id.setter
    def signal_id(self, value: str):
        raise SPyValueError("'signal_id' is read-only")

    @property
    def statistic_key(self) -> str:
        return self._get_column_definition()['statisticKey']

    @statistic_key.setter
    def statistic_key(self, value: str):
        raise SPyValueError("'statistic_key' is read-only")

    @property
    def statistic_name(self) -> str:
        """
        The name of the statistic that the column is displaying.
        """
        return self.CONSTANTS.SIGNAL_STATISTICS_TO_NAME[self.statistic_key]

    @property
    def signal_name(self):
        return self._getter_workstep.display_items.loc[
            self._getter_workstep.display_items['ID'] == self.signal_id, 'Name'].values[0]

    def __repr__(self):
        # lookup the name of the scorecard metric to show to user
        return (f"{self.signal_name} {self.statistic_name}:\n" +
                textwrap.indent("".join([
                    f"- Key: {self.key}\n",
                    f"- Type: Signal Statistic\n",
                    f"- Header: {self.header}\n" if self.header else "",
                    f"- Item ID: {self.signal_id}\n",
                    f"- Statistic: {self.statistic_name}\n",
                    f"- Grouped: {self.grouping}\n",
                    f"- Group Order: {self.group_order}\n" if self.grouping else "",
                    f"- Aggregation Function: {self.aggregation_function}\n" if self.aggregation_function else ""
                ]), prefix="    "))


ColT = TypeVar('ColT', bound=TableColumn)


class TableColumns(ABC, ContextSwitchable, MutableSequence, Generic[ColT]):
    TABLE_MODE = None

    def __init__(self, parent):
        self._parent = parent
        super().__init__(parent)

    def __repr__(self) -> str:
        columns_repr = "\n".join(f"  - {repr(column)}" for column in self)
        columns_repr = columns_repr if columns_repr else "  No columns"
        return ("Columns:\n" +
                textwrap.indent(columns_repr, prefix="  "))

    def __getitem__(self, key: Union[str, int]) -> ColT:
        # get item either by index or by key
        if isinstance(key, str):
            # if key is a string, get the column by key
            columns = self._get_interactable_columns()
            column = next((column for column in columns if column.key == key), None)
            if not column:
                raise IndexError(f"Column with key '{key}' not found")
        else:
            column = self._get_interactable_columns()[key]

        return column

    def __setitem__(self, index: int, val: TableColumn) -> None:
        raise TypeError("Adding columns should be done through the appropriate method")

    def __delitem__(self, index: int) -> None:
        column_key = self[index].key
        self._remove_column(column_key)

    def __len__(self) -> int:
        return len(self._get_interactable_columns())

    def insert(self, index: int, value: ColT):
        raise TypeError("Adding columns should be done through the appropriate method")

    @property
    def grouped_columns(self) -> List[ColT]:
        """
        Set the columns that are grouped in the table. The columns must already be present in the table.
        The order of the columns in the list will determine the order that the columns are grouped in.

        Examples
        --------
        Add column for capsule property Batch ID and group the table by this column:

        >>> workbook: spy.workbooks.Analysis
        >>> worksheet = workbook.worksheets['Worksheet Name']
        >>> batch_id_column = worksheet.condition_table_toolbar.columns.add_column_property('Batch ID')
        >>> worksheet.condition_table_toolbar.group = True # enable grouping
        >>> worksheet.condition_table_toolbar.columns.grouped_columns = [batch_id_column] # group table by asset
        """
        return self._get_grouped_columns()

    @grouped_columns.setter
    def grouped_columns(self, value: list[ColT]):

        self._set_grouped_columns(value)

    @classmethod
    def get_column_store(cls, workstep):
        return TableToolbar.get_config_for_table_mode(workstep, 'columns', cls.TABLE_MODE)

    def _add_column(self, column_definition: dict) -> TableColumn:
        column_store = self.get_column_store(self._setter_workstep)
        # if column already exists, just return the column
        if column_definition['key'] in [column['key'] for column in column_store]:
            return self._column_factory(column_definition)
        column_store.append(column_definition)
        # return column so it can be interacted with
        return self._column_factory(column_definition)

    def _remove_column(self, key: str):
        column_store = self.get_column_store(self._setter_workstep)
        column_to_remove = next(
            column for column in column_store if column['key'] == key)
        column_store.remove(column_to_remove)
        self._refresh_group_order()

    def _get_columns(self) -> List[ColT]:
        # this will get everything that is stored as a column in the table store
        # this includes columns that are actually displayed as rows
        column_definitions = self.get_column_store(self._getter_workstep)
        columns = []
        for column_definition in column_definitions:
            column = self._column_factory(column_definition)
            if column:
                columns.append(column)
        return columns

    def _set_columns(self, value: List[ColT]):
        # first check that all columns are present in the table
        # then loop through the whole column store and either rearrange the whole thing
        # ensure that all columns are present in the table
        if not all(column in self for column in value):
            raise SPyValueError('All columns must be in the table')

        new_column_store = [row._get_column_definition() for row in self._get_rows()]
        for column in value:
            new_column_store.append(column._get_column_definition())

        TableToolbar.set_config_for_table_mode(self._setter_workstep, 'columns', self.TABLE_MODE,
                                               new_column_store)
        self._refresh_group_order()

    def _get_interactable_columns(self) -> List[ColT]:
        # this will get only columns that are displayed as columns in the table
        return [column for column in self._get_columns() if not column.is_row]

    def _get_rows(self) -> List[ColT]:
        # this will get only columns that are displayed as rows in the table
        return [column for column in self._get_columns() if column.is_row]

    def _refresh_group_order(self):
        """
        Refresh the group order of all grouped columns in the table. Needs to be called when
        columns are removed to keep the group order consistent in the workstep.
        """
        self.grouped_columns = self.grouped_columns

    @abstractmethod
    def _get_grouped_columns(self) -> List[ColT]:
        pass

    @abstractmethod
    def _set_grouped_columns(self, value: List[ColT]):
        pass

    @abstractmethod
    def _column_factory(self, column_definition: dict):
        pass


class ConditionTableColumns(TableColumns[ConditionTableColumn]):
    """
    Column configuration for Condition Tables. Use methods on this object to add or remove columns from the table.
    """

    class CONSTANTS:
        DEFAULT_COLUMN_DEFINITION: Dict[str, Any] = {
            'headerTextStyle': ['bold'],
            'headerTextAlign': 'center'
        }

    TABLE_MODE = TableToolbar.CONSTANTS.CONDITION_TABLE_MODE

    def _add_column(self, *args, **kwargs) -> ConditionTableColumn:
        return cast(ConditionTableColumn, super()._add_column(*args, **kwargs))

    def _column_factory(self, column_definition: dict) -> Union[None, ConditionTableColumn]:
        # given from most to least specific
        column_type_classes = [
            ConditionTableRowCustomText,
            ConditionTableRow,
            ConditionTableColumnReservedProperty,
            ConditionTableColumnSignalStatistic,
            ConditionTableColumnScorecard,
            ConditionTableColumn
        ]
        for column_type_class in column_type_classes:
            if column_type_class.validate_definition(column_definition):
                return column_type_class(column_definition, self._parent)

    def add_column_property(self, property_name: str) -> ConditionTableColumn:
        """
        Add and return a capsule or condition property column to the table.

        Parameters
        ----------
        property_name: str
            The name of the property to add as a column

        """
        column_definition = self.CONSTANTS.DEFAULT_COLUMN_DEFINITION.copy()
        # There are four 'Condition Extra Columns' that are treated differently by the frontend
        # Start Time, End Time, Name, and Asset have special keys and do not have type 'capsuleProperty'
        # To prevent users from having to know that you have to put nameExpression to get the condition name
        # Do a lookup from the name in the frontend to the key used in the workstep.
        if property_name.lower() not in ConditionTableColumnReservedProperty.CONSTANTS.KEYS_OR_VALUES:
            column_definition['type'] = 'capsuleProperty'
        if property_name.lower() in ConditionTableColumnReservedProperty.CONSTANTS.KEY_MAPPING:
            property_name = ConditionTableColumnReservedProperty.CONSTANTS.KEY_MAPPING[property_name.lower()]
        column_definition['key'] = property_name
        return self._add_column(column_definition)

    def remove_column_property(self, property_name: str):
        """
        Remove a capsule or condition property column from the table.

        Parameters
        ----------
        property_name: str
            The name of the property to remove from the table
        """
        if property_name.lower() in ConditionTableColumnReservedProperty.CONSTANTS.KEY_MAPPING:
            property_name = ConditionTableColumnReservedProperty.CONSTANTS.KEY_MAPPING[property_name]
        self._remove_column(property_name)

    @docstring_parameter(list(TableColumn.CONSTANTS.SIGNAL_STATISTICS.keys()))
    def add_column_statistic(
            self, items, statistic: str
    ) -> Union[ConditionTableColumnSignalStatistic, List[ConditionTableColumnSignalStatistic]]:
        """
        Add and return one or multiple signal statistic columns to the table.
        All items must already be present in the details pane before they can be added as columns.
        Statistic columns can only be added for signals in condition tables.

        Parameters
        ----------
        items: Union[pd.DataFrame, pd.Series, List[Dict[str, str]], Dict[str, str], str]
            List of items to add as columns, or a single item. Can supply as:
                - pd.DataFrame or pd.Series containing 'ID' column
                - List of dictionaries with 'ID' key
                - Dictionary with 'ID' key
                - String of the item ID
        statistic: str
            Statistic to apply to each of the items. Must be one of: {0}
        """
        items: Union[pd.DataFrame, pd.Series, List[Dict[str, str]], Dict[str, str], str]
        if isinstance(items, pd.DataFrame):
            items = items.to_dict(orient='records')
        elif isinstance(items, pd.Series):
            items = [items.to_dict()]
        elif isinstance(items, list):
            items = items
        elif isinstance(items, str):
            items = [{'ID': items}]
        else:
            items = [items]
        if not items:
            raise SPyValueError("Items must be a non-empty list, DataFrame, Series, or string")

        columns_to_return = []
        for item in items:
            if not _common.present(item, 'ID'):
                raise SPyValueError(f'Item must have an ID')
            if _common.is_guid(item['ID']):
                columns_to_return.append(self._add_single_column_statistic(item['ID'], statistic))
            else:
                raise SPyValueError(f'Expecting input of ID, got {item["ID"]}')

        # if only a single column was added, return it directly to stay consistent with add_column_property
        if len(columns_to_return) == 1:
            return columns_to_return[0]
        else:
            return columns_to_return

    def _add_single_column_statistic(self, item_id: str, statistic: str) -> ConditionTableColumnSignalStatistic:
        if item_id not in self._parent.display_items['ID'].values:
            # item must exist in details pane before it can be added
            raise SPyValueError(f"Item with ID {item_id} must be added to the details pane before adding as a column")
        # make sure it's a signal!
        if self._parent.display_items.loc[self._parent.display_items['ID'] == item_id, 'Type'].values[0] != 'Signal':
            raise SPyValueError(f"Item with ID {item_id} is not a signal.")

        if statistic.lower() not in TableColumn.CONSTANTS.SIGNAL_STATISTICS:
            raise SPyValueError(f"'statistic' must be one of {list(TableColumn.CONSTANTS.SIGNAL_STATISTICS.keys())}")

        column_definition = self.CONSTANTS.DEFAULT_COLUMN_DEFINITION.copy()

        statistic_key = TableColumn.CONSTANTS.SIGNAL_STATISTICS[statistic.lower()]
        column_definition['key'] = f"{statistic_key}_{item_id}"
        column_definition['statisticKey'] = statistic_key
        column_definition['signalId'] = item_id
        return cast(ConditionTableColumnSignalStatistic, self._add_column(column_definition))

    def remove_column_statistic(self, item_id: str, statistic: str):
        """
        Remove a signal statistic column from the table.

        Parameters
        ----------
        item_id: str
            ID of the signal to remove from the table
        statistic: str
            Statistic to remove for the signal. Must be one of: {0}
        """
        if statistic.lower() not in TableColumn.CONSTANTS.SIGNAL_STATISTICS:
            raise SPyValueError(f"Statistic must be one of {TableColumn.CONSTANTS.SIGNAL_STATISTICS.keys()}")
        statistic_key = TableColumn.CONSTANTS.SIGNAL_STATISTICS[statistic.lower()]
        key = f"{statistic_key}_{item_id}"
        self._remove_column(key)

    @docstring_parameter(ConditionTableToolbar.add_row_custom_text.__doc__)
    def add_row_custom_text(self, column_values: Union[List[str], Dict[str, str]]) -> ConditionTableRowCustomText:
        """
        Add and return a row that displays custom text in the table.
        Once returned, the row can be manipulated to adjust the appearance of the custom text.

        Parameters
        ----------
        column_values: Union[List[str], Dict[str, str]]
            List of string values to display in the row or a dictionary of column keys and values

        """
        cells = {}
        if isinstance(column_values, list):
            # if given a list of strings, assign to columns in the order that they are currently arranged
            if len(column_values) > len(self):
                raise SPyValueError(f'{len(column_values)} column values provided, but only '
                                    f'{len(self)} '
                                    f'columns exist')
            for i, value in enumerate(column_values):
                cells[self[i].key] = value
        if isinstance(column_values, dict):
            cells = column_values
        column_definition = self.CONSTANTS.DEFAULT_COLUMN_DEFINITION.copy()
        column_definition['type'] = 'text'
        column_definition['key'] = _common.new_placeholder_guid()
        column_definition['cells'] = cells
        return cast(ConditionTableRowCustomText, self._add_column(column_definition))

    @docstring_parameter(ConditionTableToolbar.remove_row_custom_text.__doc__)
    def remove_row_custom_text(self):
        """
        Remove all rows that displays custom text in the table.
        """
        columns_to_remove = [column for column in self._get_columns() if
                             isinstance(column, ConditionTableRowCustomText)]
        for column in columns_to_remove:
            self._remove_column(column.key)

    @docstring_parameter(ConditionTableToolbar.add_row_column_labels.__doc__)
    def add_row_column_labels(self) -> ConditionTableRow:
        """
        Enable and return the row that displays column labels in the table.
        Once returned, the row can be manipulated to adjust the appearance of the column labels.
        """
        column_definition = self.CONSTANTS.DEFAULT_COLUMN_DEFINITION.copy()
        column_definition['key'] = ConditionTableRow.CONSTANTS.COLUMN_LABELS
        return cast(ConditionTableRow, self._add_column(column_definition))

    @docstring_parameter(ConditionTableToolbar.remove_row_column_labels.__doc__)
    def remove_row_column_labels(self):
        """
        Hide the row that displays column labels in the table.
        """
        if ConditionTableRow.CONSTANTS.COLUMN_LABELS in self._get_columns():
            self._remove_column(ConditionTableRow.CONSTANTS.COLUMN_LABELS)

    @docstring_parameter(ConditionTableToolbar.add_row_unit_of_measure.__doc__)
    def add_row_unit_of_measure(self) -> ConditionTableRow:
        """
        Enable and return the row that displays units of measure in the table.
        When the row is present, units will be moved from within each cell to the row.
        Once returned, this row can be manipulated to adjust the appearance of the units of measure.
        """
        column_definition = self.CONSTANTS.DEFAULT_COLUMN_DEFINITION.copy()
        column_definition['key'] = ConditionTableRow.CONSTANTS.UNIT_OF_MEASURE
        return cast(ConditionTableRow, self._add_column(column_definition))

    @docstring_parameter(ConditionTableToolbar.remove_row_unit_of_measure.__doc__)
    def remove_row_unit_of_measure(self):
        """
        Hide the row that displays units of measure in the table.
        """
        if ConditionTableRow.CONSTANTS.UNIT_OF_MEASURE in self._get_columns():
            self._remove_column(ConditionTableRow.CONSTANTS.UNIT_OF_MEASURE)

    def _get_grouped_columns(self) -> List[TableColumn]:
        return sorted([column for column in self if column.grouping], key=lambda column: column.group_order)

    def _set_grouped_columns(self, value: List[TableColumn]):
        if not (isinstance(value, list)) and not (isinstance(value, ConditionTableColumns)):
            raise SPyTypeError(f"'grouped_columns' must be a list of columns")

        # ensure all columns given are currently in the table
        if not all(column in self for column in value):
            raise SPyValueError('All columns provided must be in the table')

        # do this in two stages. first remove grouping from all columns
        # and then apply to the columns specified
        for column in self:
            column.grouping = False
            column.group_order = None

        if value:
            for row_group_order, column in enumerate(value):
                column.grouping = True
                column.group_order = row_group_order


class SimpleTableColumns(TableColumns):
    """
    Column configuration for Simple Tables. Use methods on this object to add or remove columns from the table.
    """
    TABLE_MODE = TableToolbar.CONSTANTS.SIMPLE_TABLE_MODE

    def _column_factory(self, column_definition: dict):
        raise NotImplementedError('SPy does not currently support manipulating columns in simple table mode')

    def add_column_property(self, key: str):
        raise NotImplementedError('SPy does not currently support manipulating columns in simple table mode')

    def remove_column_property(self, key: str):
        raise NotImplementedError('SPy does not currently support manipulating columns in simple table mode')

    def add_column_custom_text(self, column_values: Union[List[str], Dict[str, str]]):
        raise NotImplementedError('SPy does not currently support manipulating columns in simple table mode')

    def remove_column_custom_text(self):
        raise NotImplementedError('SPy does not currently support manipulating columns in simple table mode')

    def _get_interactable_columns(self) -> List[TableColumn]:
        raise NotImplementedError('SPy does not currently support manipulating columns in simple table mode')

    def _get_grouped_columns(self) -> List[TableColumn]:
        raise NotImplementedError('SPy does not currently support manipulating columns in simple table mode')

    def _set_grouped_columns(self, value: List[TableColumn]):
        raise NotImplementedError('SPy does not currently support manipulating columns in simple table mode')
